#!/usr/bin/env python3
# -*- coding: utf-8 -*-
# ================================================== #
# This file is a part of PYGPT package               #
# Website: https://pygpt.net                         #
# GitHub:  https://github.com/szczyglis-dev/py-gpt   #
# MIT License                                        #
# Created By  : Marcin Szczygliński                  #
# Updated Date: 2025.07.16 15:00:00                  #
# ================================================== #

import os

from pygpt_net.core.render.web.syntax_highlight import SyntaxHighlight


class Body:

    def __init__(self, window=None):
        """
        HTML Body

        :param window: Window instance
        """
        self.window = window
        self.highlight = SyntaxHighlight(window)

    def prepare_styles(self) -> str:
        """
        Prepare CSS styles

        :return: CSS styles
        """
        syntax_dark = [
            "dracula",
            "fruity",
            "github-dark",
            "gruvbox-dark",
            "inkpot",
            "material",
            "monokai",
            "native",
            "nord",
            "nord-darker",
            "one-dark",
            "paraiso-dark",
            "rrt",
            "solarized-dark",
            "stata-dark",
            "vim",
            "zenburn",
        ]
        syntax_style = self.window.core.config.get("render.code_syntax")
        if syntax_style is None or syntax_style == "":
            syntax_style = "default"
        fonts_path = os.path.join(self.window.core.config.get_app_path(), "data", "fonts").replace("\\", "/")
        # loader
        stylesheet = """
        .lds-ring {
          /* change color here */
          color: #1c4c5b
        }
        .lds-ring,
        .lds-ring div {
          box-sizing: border-box;
        }
        .lds-ring {
          display: inline-block;
          position: relative;
          width: 80px;
          height: 80px;
        }
        .lds-ring div {
          box-sizing: border-box;
          display: block;
          position: absolute;
          width: 64px;
          height: 64px;
          margin: 8px;
          border: 8px solid currentColor;
          border-radius: 50%;
          animation: lds-ring 1.2s cubic-bezier(0.5, 0, 0.5, 1) infinite;
          border-color: currentColor transparent transparent transparent;
        }
        .lds-ring div:nth-child(1) {
          animation-delay: -0.45s;
        }
        .lds-ring div:nth-child(2) {
          animation-delay: -0.3s;
        }
        .lds-ring div:nth-child(3) {
          animation-delay: -0.15s;
        }
        @keyframes lds-ring {
          0% {
            transform: rotate(0deg);
          }
          100% {
            transform: rotate(360deg);
          }
        }
        """
        stylesheet += self.window.controller.theme.markdown.get_web_css().replace('%fonts%', fonts_path)
        if syntax_style in syntax_dark:
            stylesheet += "pre { color: #fff; }"
        else:
            stylesheet += "pre { color: #000; }"

        stylesheet += """
          body {max-width: 100%; } 
          pre { margin-top: 0; margin-bottom: 0.25rem; } 
          a:hover { cursor: pointer; }
          .output-image { max-width: 100%; height: auto; }
        """
        return stylesheet + " " + self.highlight.get_style_defs()

    def get_html(self, pid: int) -> str:
        """
        Build webview HTML code

        :return: HTML code
        """
        classes = []
        classes_str = ""
        style = self.window.core.config.get("theme.style", "blocks")
        if self.window.core.config.get('render.blocks'):
            classes.append("display-blocks")
        if self.window.core.config.get('ctx.edit_icons'):
            classes.append("display-edit-icons")
        classes.append("theme-" + style)
        if classes:
            classes_str = ' class="' + " ".join(classes) + '"'

        return """
        <!DOCTYPE html>
        <html>
        <head>
            <style>
                """ + self.prepare_styles() + """
            </style>
        <script type="text/javascript" src="qrc:///qtwebchannel/qwebchannel.js"></script>
        <script type="text/javascript" src="qrc:///js/highlight.min.js"></script>
        <script type="text/javascript" src="qrc:///js/auto-render.min.js"></script>
        <script>

        let scrollTimeout = null;
        let prevScroll = 0;
        let bridge;
        let pid = """ + str(pid) + """
        new QWebChannel(qt.webChannelTransport, function (channel) {
            bridge = channel.objects.bridge;
        });
        history.scrollRestoration = "manual";
        document.addEventListener('keydown', function(event) {
            if (event.ctrlKey && event.key === 'f') {
                //window.location.href = 'bridge://open_find:' + pid; // send to bridge
                event.preventDefault();
            }
            if (event.key === 'Escape') {
                window.location.href = 'bridge://escape'; // send to bridge
                event.preventDefault();
            }
        });
        document.addEventListener('click', function(event) {
            if (event.target.tagName !== 'A' && !event.target.closest('a')) {
                window.location.href = 'bridge://focus';
            }
        });
        function log(text) {
            if (bridge) {
                bridge.log(text);
            }
        }
        function sanitize(content) {
            var parser = new DOMParser();
            var doc = parser.parseFromString(content, "text/html");
            var codeElements = doc.querySelectorAll('code, pre');
            codeElements.forEach(function(element) {
                var html = element.outerHTML;
                var newHtml = html.replace(/&amp;lt;/g, '&lt;').replace(/&amp;gt;/g, '&gt;');
                element.outerHTML = newHtml;
            });
            return doc.documentElement.outerHTML;
        }
        function highlightCode() {
            document.querySelectorAll('pre code').forEach(el => {
                if (!el.classList.contains('hljs')) hljs.highlightElement(el);
            });      
            restoreCollapsedCode();   
        }
        function scrollToBottom() {
            getScrollPosition();  // store using bridge
            if (scrollTimeout !== null) {
                return;
            }
            if (document.body.scrollHeight > prevScroll) {
                scrollTimeout = setTimeout(function() {
                    window.scrollTo(0, document.body.scrollHeight);
                    prevScroll = document.body.scrollHeight;
                    getScrollPosition();  // store using bridge
                    scrollTimeout = null;
                }, 30);
            }
        }
        
        // ----------------------------------
        function insertOutput(content, type) {
            const element = document.getElementById('_append_output_');
            if (element) {
                // first remove all nodes inside
                const pre = document.createElement('pre');
                pre.classList.add('type-' + type);
                const code = document.createElement('code');
                code.classList.add('language-python');
                code.innerHTML = sanitize(content);
                pre.appendChild(code);
                element.appendChild(pre);
            }
            highlightCode();
            scrollToBottom();
        }
        function replaceOutput(content) {
            const element = document.getElementById('_append_output_');
            if (element) {
                // first remove all nodes inside
                element.innerHTML = '';
                const pre = document.createElement('pre');
                const code = document.createElement('code');
                code.classList.add('language-python');
                code.innerHTML = sanitize(content);
                pre.appendChild(code);
                element.appendChild(pre);
            }
            highlightCode();
            scrollToBottom();
        }
        function beginOutput(type) {
            const element = document.getElementById('_append_output_');
            if (element) {
                const pre = document.createElement('pre');
                pre.classList.add('type-' + type);
                const code = document.createElement('code');
                code.classList.add('language-python');
                pre.appendChild(code);
                element.appendChild(pre);
            }
        }
        function endOutput(type) {
            const element = document.getElementById('_append_output_');
            if (element) {
               // get last <pre> element, if empty then remove it
                const nodes = element.querySelectorAll('pre code');
                if (nodes.length > 0) {
                    const last = nodes[nodes.length - 1];
                    if (last.innerHTML.trim() === '') {
                        last.parentElement.remove(); // remove empty <pre>
                    }
                }
            }
            highlightCode()
        }
        function appendToOutput(content) {
            const element = document.getElementById('_append_output_');
            if (element) {
                const nodes = element.querySelectorAll('pre code');
                if (nodes.length === 0) {
                    appendBegin();
                    const pre = document.querySelector('pre');
                    const code = pre.querySelector('code');
                    code.innerHTML = content;
                } else {
                    const last = nodes[nodes.length - 1];
                    last.innerHTML += content;
                }
            }
            scrollToBottom();
        }
        function appendImage(path, url) {
            const element = document.getElementById('_append_output_');
            if (element) {
                const a = document.createElement('a');
                a.href = url || path;  // use provided URL or local path
                const img = document.createElement('img');
                img.src = path;
                img.classList.add('output-image');
                img.alt = 'Output Image';
                img.title = 'Output Image';
                a.appendChild(img);
                element.appendChild(a);
            }
        }
        
        function clearOutput() {
            const element = document.getElementById('_append_output_');
            if (element) {
                element.textContent = '';
            }
        }
        function updateCSS(styles) {
            const style = document.createElement('style');
            style.innerHTML = styles;
            const oldStyle = document.querySelector('style');
            if (oldStyle) {
                oldStyle.remove();
            }
            document.head.appendChild(style);
        }
        function bridgeCopyCode(text) {
            if (bridge) {
                bridge.copy_text(text);
            }
        }
        function bridgePreviewCode(text) {
            if (bridge) {
                bridge.preview_text(text);
            }
        }
        function bridgeRunCode(text) {
            if (bridge) {
                bridge.run_text(text);
            }
        }
        function bridgeUpdateScrollPosition(pos) {
            if (bridge) {
                bridge.update_scroll_position(pos);
            }
        }
        function getScrollPosition() {
            bridgeUpdateScrollPosition(window.scrollY);
        }
        function setScrollPosition(pos) {
            window.scrollTo(0, pos);
            prevScroll = parseInt(pos);
        }  
        document.addEventListener('DOMContentLoaded', function() {
            const container = document.getElementById('container');
            container.addEventListener('click', function(event) {                
                // btn collapse
                const collapseButton = event.target.closest('.code-header-collapse');
                if (collapseButton) {
                    event.preventDefault();
                    const parent = collapseButton.closest('.code-wrapper');
                    const index = parent.getAttribute('data-index');
                    const source = parent.querySelector('code');
                    const localeCollapse = parent.getAttribute('data-locale-collapse');
                    const localeExpand = parent.getAttribute('data-locale-expand');
                    if (source) {
                        if (source.style.display === 'none') {
                            source.style.display = 'block';
                            collapseButton.classList.remove('collapsed');
                            const idx = collapsed_idx.indexOf(index);
                            if (idx !== -1) {
                                collapsed_idx.splice(idx, 1);
                                const collapseSpan = collapseButton.querySelector('span');
                                if (collapseSpan) {
                                    collapseSpan.textContent = localeCollapse;
                                }
                            }
                        } else {
                            source.style.display = 'none';
                            collapseButton.classList.add('collapsed');
                            if (!collapsed_idx.includes(index)) {
                                collapsed_idx.push(index);
                                const collapseSpan = collapseButton.querySelector('span');
                                if (collapseSpan) {
                                    collapseSpan.textContent = localeExpand;
                                }
                            }
                        }
                    }
                }
            });
        });
        </script>
        </head>
        <body """ + classes_str + """>
        <div id="container">
            <div id="_append_output_" class="append_output"></div>    
        </div>
        </body>
        </html>
        """



#!/usr/bin/env python3
# -*- coding: utf-8 -*-
# ================================================== #
# This file is a part of PYGPT package               #
# Website: https://pygpt.net                         #
# GitHub:  https://github.com/szczyglis-dev/py-gpt   #
# MIT License                                        #
# Created By  : Marcin Szczygliński                  #
# Updated Date: 2025.06.29 18:00:00                  #
# ================================================== #

from PySide6.QtWidgets import QSizePolicy, QLayout
from PySide6.QtCore import QPoint, QRect, QSize, Qt

class FlowLayout(QLayout):
    def __init__(self, parent=None, spacing=5):
        super().__init__(parent)
        self.setSpacing(spacing)
        self.itemList = []

    def addItem(self, item):
        self.itemList.append(item)

    def count(self):
        return len(self.itemList)

    def itemAt(self, index):
        if 0 <= index < len(self.itemList):
            return self.itemList[index]
        return None

    def takeAt(self, index):
        if 0 <= index < len(self.itemList):
            return self.itemList.pop(index)
        return None

    def expandingDirections(self):
        return Qt.Orientations(0)

    def hasHeightForWidth(self):
        return True

    def heightForWidth(self, width):
        return self.doLayout(QRect(0, 0, width, 0), True)

    def setGeometry(self, rect):
        super().setGeometry(rect)
        self.doLayout(rect, False)

    def sizeHint(self):
        return self.minimumSize()

    def minimumSize(self):
        size = QSize()
        for item in self.itemList:
            size = size.expandedTo(item.minimumSize())
        margins = self.contentsMargins()
        size += QSize(margins.left() + margins.right(), margins.top() + margins.bottom())
        return size

    def doLayout(self, rect, testOnly):
        x = rect.x()
        y = rect.y()
        lineHeight = 0
        for item in self.itemList:
            wid = item.widget()
            spaceX = self.spacing() + wid.style().layoutSpacing(
                QSizePolicy.PushButton, QSizePolicy.PushButton, Qt.Horizontal)
            spaceY = self.spacing() + wid.style().layoutSpacing(
                QSizePolicy.PushButton, QSizePolicy.PushButton, Qt.Vertical)
            nextX = x + item.sizeHint().width() + spaceX
            if nextX - spaceX > rect.right() and lineHeight > 0:
                x = rect.x()
                y = y + lineHeight + spaceY
                nextX = x + item.sizeHint().width() + spaceX
                lineHeight = 0
            if not testOnly:
                item.setGeometry(QRect(QPoint(x, y), item.sizeHint()))
            x = nextX
            lineHeight = max(lineHeight, item.sizeHint().height())
        return y + lineHeight - rect.y()


#!/usr/bin/env python3
# -*- coding: utf-8 -*-
# ================================================== #
# This file is a part of PYGPT package               #
# Website: https://pygpt.net                         #
# GitHub:  https://github.com/szczyglis-dev/py-gpt   #
# MIT License                                        #
# Created By  : Marcin Szczygliński                  #
# Updated Date: 2025.07.11 01:00:00                  #
# ================================================== #

import os
import re
from typing import Optional, List, Dict

from llama_index.core.schema import ImageDocument
from llama_index.core.llms import (
    ChatMessage,
    ImageBlock,
    TextBlock,
    MessageRole,
)

from pygpt_net.item.attachment import AttachmentItem
from pygpt_net.item.ctx import CtxItem
from pygpt_net.utils import is_image


class Context:
    def __init__(self, window=None):
        """
        Context core

        :param window: Window instance
        """
        self.window = window
        self.attachments = {}
        self.urls = []

    def get_messages(
            self,
            input_prompt: str,
            system_prompt: str,
            history: Optional[List[CtxItem]] = None,
            multimodal: bool = False,
            prev_message = None,
            allow_native_tool_calls: bool = False,
            attachments: Dict[str, AttachmentItem] = None,
    ):
        """
        Get messages from db

        :param input_prompt: input prompt
        :param system_prompt: system prompt
        :param history: history
        :param multimodal: multimodal flag
        :param prev_message: previous message
        :param allow_native_tool_calls: allow native tool calls
        :param attachments: attachments
        :return: Messages
        """
        messages = []

        # tokens config
        model = self.window.core.config.get('model')
        model_id = self.window.core.models.get_id(model)
        mode = self.window.core.config.get('mode')

        used_tokens = self.window.core.tokens.from_user(input_prompt, system_prompt)  # threshold and extra included
        max_tokens = self.window.core.config.get('max_total_tokens')
        model_ctx = self.window.core.models.get_num_ctx(model_id)

        # fit to max model tokens
        if max_tokens > model_ctx:
            max_tokens = model_ctx

        if self.window.core.config.get('use_context'):
            items = self.window.core.ctx.get_history(
                history,
                model_id,
                mode,
                used_tokens,
                max_tokens,
            )
            for item in items:
                # input
                if item.final_input is not None and item.final_input != "":
                    messages.append(ChatMessage(
                        role=MessageRole.USER,
                        content=item.final_input
                    ))
                # output
                if item.final_output is not None and item.final_output != "":
                    msg = ChatMessage(
                        role=MessageRole.ASSISTANT,
                        content=item.final_output
                    )
                    messages.append(msg)

                # ---- if tool output ----
                is_last_item = item == items[-1] if items else False
                if (is_last_item
                        and prev_message
                        and allow_native_tool_calls
                        and item.extra
                        and isinstance(item.extra, dict)):
                    if "tool_calls" in item.extra and isinstance(item.extra["tool_calls"], list):
                        for tool_call in item.extra["tool_calls"]:
                            if "function" in tool_call:
                                if "id" not in tool_call or "name" not in tool_call["function"]:
                                    continue
                                if tool_call["id"] and tool_call["function"]["name"]:
                                    if "tool_output" in item.extra and isinstance(item.extra["tool_output"], list):
                                        for tool_output in item.extra["tool_output"]:
                                            if ("cmd" in tool_output
                                                    and tool_output["cmd"] == tool_call["function"]["name"]):
                                                last_msg = messages[-1] if messages else None
                                                if last_msg and last_msg.role == "assistant":
                                                    if prev_message:
                                                        last_msg = prev_message  # prev message with tool calls
                                                        messages[-1] = last_msg

                                                msg = ChatMessage(
                                                    role=MessageRole.TOOL,
                                                    content=str(tool_output),
                                                    additional_kwargs={"tool_call_id": tool_call["id"]}
                                                )
                                                messages.append(msg)
                                                break
                                            elif "result" in tool_output:
                                                # if result is present, append it as function call output
                                                last_msg = messages[-1] if messages else None
                                                if last_msg and last_msg.role == "assistant":
                                                    if prev_message:
                                                        last_msg = prev_message  # prev message with tool calls
                                                        messages[-1] = last_msg

                                                msg = ChatMessage(
                                                    role=MessageRole.TOOL,
                                                    content=str(tool_output["result"]),
                                                    additional_kwargs={"tool_call_id": tool_call["id"]}
                                                )
                                                messages.append(msg)
                                                break


        return messages

    def add_user(
            self,
            query: str,
            attachments: Dict[str, AttachmentItem] = None) -> ChatMessage:
        """
        Add user message

        :param query: input query
        :param attachments: attachments
        """
        blocks = []
        blocks.append(
            TextBlock(text=query)
        )

        self.attachments = {}  # reset attachments, only current prompt
        self.urls = []

        #https://pygpt.net/assets/img/img3.jpg?v=2024-11-28

        # extract URLs from prompt
        urls = self.extract_urls(query)
        if len(urls) > 0:
            for url in urls:
                blocks.append(
                    ImageBlock(url=url)
                )
                self.urls.append(url)

        # if attachments are provided, add them to blocks
        if attachments is not None and len(attachments) > 0:
            for id in attachments:
                attachment = attachments[id]
                if os.path.exists(attachment.path):
                    if is_image(attachment.path):
                        blocks.append(
                            ImageBlock(path=attachment.path)
                        )
                        self.attachments[id] = attachment.path

        msg = ChatMessage(
            role=MessageRole.USER,
            blocks=blocks,
        )
        return msg



        urls.append(attachment.path)
        msg = ChatMessage(
            role=MessageRole.USER,
            content=query,
        )
        if urls:
            image_documents = [ImageDocument(image_path=url) for url in urls]
            if image_documents:
                msg = ChatMessage(
                    role=MessageRole.USER,
                    content=query,
                    image_documents=image_documents,
                )
        return msg

    def add_system(self, prompt: str) -> ChatMessage:
        """
        Add system message to db

        :param prompt: system prompt
        """
        return ChatMessage(
            role=MessageRole.SYSTEM,
            content=prompt,
        )

    def append_images(self, ctx: CtxItem):
        """
        Append images content to context item

        :param ctx: context
        """
        images = self.get_attachments()  # dict -> key: id, value: path
        urls = self.get_urls()  # list

        # store sent images in ctx
        if len(images) > 0:
            ctx.images = self.window.core.filesystem.make_local_list(list(images.values()))
        if len(urls) > 0:
            ctx.images = urls
            ctx.urls = urls

    def extract_urls(self, text: str) -> List[str]:
        """
        Extract img urls from text

        :param text: text
        :return: list of img urls
        """
        urls = re.findall(r'(https?://\S+)', text)
        img_urls = []
        for url in urls:
            if self.is_image(url):
                img_urls.append(url)

        # if single url, return it
        if "".join(urls).strip() == text.strip():
            return urls

        return img_urls

    def is_image(self, path: str) -> bool:
        """
        Check if url is image

        :param path: url
        :return: True if image
        """
        return path.lower().endswith(('.png', '.jpg', '.jpeg', '.tiff', '.bmp', '.gif', '.webp'))

    def get_attachments(self) -> Dict[str, str]:
        """
        Get attachments

        :return: attachments dict
        """
        return self.attachments

    def get_urls(self) -> List[str]:
        """
        Get urls

        :return: urls list
        """
        return self.urls



#!/usr/bin/env python3
# -*- coding: utf-8 -*-
# ================================================== #
# This file is a part of PYGPT package               #
# Website: https://pygpt.net                         #
# GitHub:  https://github.com/szczyglis-dev/py-gpt   #
# MIT License                                        #
# Created By  : Marcin Szczygliński                  #
# Updated Date: 2024.12.14 22:00:00                  #
# ================================================== #

import json
import os
import uuid
from typing import Dict

from packaging.version import Version

from pygpt_net.provider.core.prompt.base import BaseProvider
from pygpt_net.item.prompt import PromptItem


class JsonFileProvider(BaseProvider):
    def __init__(self, window=None):
        super(JsonFileProvider, self).__init__(window)
        self.window = window
        self.id = "json_file"
        self.type = "prompt"
        self.config_file = 'prompts.json'

    def create_id(self) -> str:
        """
        Create unique uuid

        :return: uuid
        """
        return str(uuid.uuid4())

    def create(self, prompt: PromptItem) -> str:
        """
        Create new and return its ID

        :param prompt: PromptItem
        :return: attachment ID
        """
        if prompt.id is None or prompt.id == "":
            prompt.id = self.create_id()
        return prompt.id

    def load(self) -> Dict[str, PromptItem]:
        """
        Load prompts from file

        :return: dict
        """
        path = os.path.join(self.window.core.config.path, self.config_file)
        items = {}
        try:
            if os.path.exists(path):
                with open(path, 'r', encoding="utf-8") as file:
                    data = json.load(file)
                    if data == "" or data is None or 'items' not in data:
                        return {}
                    # deserialize
                    items = {}
                    for id in data['items']:
                        item = data['items'][id]
                        prompt = PromptItem()
                        self.deserialize(item, prompt)
                        items[id] = prompt
        except Exception as e:
            self.window.core.debug.log(e)
            items = {}

        return items

    def save(self, items: Dict[str, PromptItem]):
        """
        Save prompts to file
        """
        try:
            # update prompts
            path = os.path.join(self.window.core.config.path, self.config_file)
            data = {}
            ary = {}

            # serialize
            ary = {}
            for id in items:
                prompt = items[id]
                ary[id] = self.serialize(prompt)

            data['__meta__'] = self.window.core.config.append_meta()
            data['items'] = ary
            dump = json.dumps(data, indent=4)
            with open(path, 'w', encoding="utf-8") as f:
                f.write(dump)

        except Exception as e:
            self.window.core.debug.log(e)
            print("Error while saving prompts: {}".format(str(e)))

    def remove(self, id: str):
        """
        Delete by id

        :param id: id
        """
        pass

    def truncate(self, mode: str):
        """
        Delete all

        :param mode: mode
        """
        path = os.path.join(self.window.core.config.path, self.config_file)
        data = {'__meta__': self.window.core.config.append_meta(), 'items': {}}
        try:
            dump = json.dumps(data, indent=4)
            with open(path, 'w', encoding="utf-8") as f:
                f.write(dump)
        except Exception as e:
            self.window.core.debug.log(e)

    def patch(self, version: Version) -> bool:
        """
        Migrate presets to current app version

        :param version: current app version
        :return: True if migrated
        """
        return False

    @staticmethod
    def serialize(prompt: PromptItem) -> Dict[str, str]:
        """
        Serialize item to dict

        :return: serialized item
        """
        return {
            'id': prompt.id,
            'name': prompt.name,
            'content': prompt.content,
        }

    @staticmethod
    def deserialize(data: Dict[str, str], prompt: PromptItem):
        """
        Deserialize item from dict

        :param data: serialized item
        :param prompt: PromptItem
        """
        if 'id' in data:
            prompt.id = data['id']
        if 'name' in data:
            prompt.name = data['name']
        if 'content' in data:
            prompt.content = data['content']

    def dump(self, item: PromptItem) -> str:
        """
        Dump to string

        :param item: item to dump
        :return: dumped item as string (json)
        """
        return json.dumps(self.serialize(item))



#!/usr/bin/env python3
# -*- coding: utf-8 -*-
# ================================================== #
# This file is a part of PYGPT package               #
# Website: https://pygpt.net                         #
# GitHub:  https://github.com/szczyglis-dev/py-gpt   #
# MIT License                                        #
# Created By  : Marcin Szczygliński                  #
# Updated Date: 2025.06.23 19:00:00                  #
# ================================================== #

import copy
import os

from PySide6 import QtWidgets
from PySide6.QtCore import QTimer, Signal, Slot, QThreadPool, QEvent, Qt, QLoggingCategory
from PySide6.QtGui import QShortcut, QKeySequence
from PySide6.QtWidgets import QApplication, QMainWindow
from qt_material import QtStyleTools

from pygpt_net.core.events import BaseEvent, KernelEvent, ControlEvent
from pygpt_net.container import Container
from pygpt_net.controller import Controller
from pygpt_net.tools import Tools
from pygpt_net.ui import UI
from pygpt_net.utils import get_app_meta


class MainWindow(QMainWindow, QtStyleTools):

    # states
    STATE_IDLE = 'idle'
    STATE_BUSY = 'busy'
    STATE_ERROR = 'error'

    # signals
    statusChanged = Signal(str)
    stateChanged = Signal(str)
    logger_message = Signal(object)
    idx_logger_message = Signal(object)

    def __init__(self, app: QApplication, args: dict = None):
        """
        Main window

        :param app: QApplication instance
        :param args: launcher arguments
        """
        super().__init__()
        self.app = app
        self.args = args
        self.timer = None
        self.post_timer = None
        self.update_timer = None
        self.threadpool = None
        self.is_closing = False
        self.timer_interval = 30
        self.post_timer_interval = 1000
        self.update_timer_interval = 300000  # check every 5 minutes
        self.state = self.STATE_IDLE
        self.prevState = None

        # load version info
        self.meta = get_app_meta()

        # setup service container
        self.core = Container(self)
        self.core.init()
        self.core.patch()  # patch version if needed
        self.core.post_setup()

        # before render, handle engine args
        self.handle_engine_args()

        # setup thread pool
        self.threadpool = QThreadPool()

        # setup controller
        self.controller = Controller(self)

        # setup tools
        self.tools = Tools(self)

        # init, load settings options, etc.
        self.controller.init()

        # setup UI
        self.ui = UI(self)
        self.ui.init()

        # global shortcuts
        self.shortcuts = []

        # setup signals
        self.statusChanged.connect(self.update_status)
        self.stateChanged.connect(self.update_state)

    def handle_engine_args(self):
        """Handle launcher arguments"""
        render_debug = False
        if self.args is not None:
            if "debug" in self.args and (self.args["debug"] == "1" or self.args["debug"] == "2"):
                render_debug = True
            if "legacy" in self.args and self.args["legacy"] == "1":
                self.core.config.set("render.engine", "legacy")
            if "disable-gpu" in self.args and self.args["disable-gpu"] == "1":
                self.core.config.set("render.open_gl", False)

        # from log level
        if self.core.config.get("log.level") in ["debug", "info"]:
            render_debug = True

        # disable/enable logging web engine context
        if not render_debug:
            log = QLoggingCategory("qt.webenginecontext")
            log.setFilterRules("*.info=false")
        else:
            os.environ["QTWEBENGINE_CHROMIUM_FLAGS"] = "--enable-logging --log-level=0"

        # OpenGL disable
        if self.core.config.get("render.open_gl") is False:
            if "QTWEBENGINE_CHROMIUM_FLAGS" in os.environ:
                os.environ["QTWEBENGINE_CHROMIUM_FLAGS"] += " --disable-gpu"
            else:
                os.environ["QTWEBENGINE_CHROMIUM_FLAGS"] = "--disable-gpu"

    def add_plugin(self, plugin):
        """
        Add a plugin to the app

        :param plugin: plugin instance
        """
        self.core.plugins.register(plugin)

    def add_llm(self, llm):
        """
        Add a Langchain LLM wrapper to the app

        :param llm: LLM wrapper instance
        """
        self.core.llm.register(llm.id, llm)

    def add_vector_store(self, store):
        """
        Add a vector store provider to the app

        :param store: Vector store provider instance
        """
        self.core.idx.storage.register(store.id, store)

    def add_loader(self, loader):
        """
        Add a data loader to the app

        :param loader: data loader instance
        """
        self.core.idx.indexing.register_loader(loader)

    def add_agent(self, agent):
        """
        Add an agent to the app

        :param agent: agent instance
        """
        self.core.agents.provider.register(agent.id, agent)

    def add_audio_input(self, provider):
        """
        Add an audio input provider to the app

        :param provider: audio input provider instance
        """
        self.core.audio.register(provider, "input")

    def add_audio_output(self, provider):
        """
        Add an audio output provider to the app

        :param provider: audio output provider instance
        """
        self.core.audio.register(provider, "output")

    def add_web(self, provider):
        """
        Add a web provider to the app

        :param provider: web provider instance
        """
        self.core.web.register(provider)

    def add_tool(self, tool):
        """
        Add a tool to the app

        :param tool: tool instance
        """
        self.tools.register(tool)

    def setup(self):
        """Setup app"""
        self.tools.setup()
        self.controller.setup()
        self.controller.plugins.setup()
        self.controller.post_setup()

    def post_setup(self):
        """Called after setup"""
        self.controller.layout.post_setup()
        self.timer = QTimer()
        self.timer.timeout.connect(self.update)
        self.timer.start(self.timer_interval)
        self.post_timer = QTimer()
        self.post_timer.timeout.connect(self.post_update)
        self.post_timer.start(self.post_timer_interval)
        self.update_timer = QTimer()
        self.update_timer.timeout.connect(self.core.updater.run_check)
        self.update_timer.start(self.update_timer_interval)
        self.logger_message.connect(self.controller.debug.handle_log)
        self.ui.post_setup()
        self.tools.post_setup()

    def update(self):
        """Called on every update (real-time)"""
        self.controller.on_update()
        self.controller.plugins.on_update()
        self.tools.on_update()

    def post_update(self):
        """Called on post-update (lazy)"""
        self.controller.debug.on_post_update()
        self.controller.plugins.on_post_update()
        self.tools.on_post_update()

    @Slot(str)
    def update_status(self, message: str = ""):
        """
        Update global status

        :param message: status message
        """
        self.dispatch(KernelEvent(KernelEvent.STATUS, {"status": str(message)}))

    @Slot(str)
    def update_state(self, state: str):
        """
        Update state

        :param state: state
        """
        self.state = state
        self.ui.tray.set_icon(state)

    @Slot(object)
    def dispatch(self, event: BaseEvent, all: bool = False):
        """
        Dispatch App event

        :param event
        :param all: True to dispatch to all plugins
        """
        self.core.dispatcher.dispatch(event, all=all)

    def closeEvent(self, event):
        """
        Handle close event

        :param event: close event
        """
        if self.core.config.get('layout.tray.minimize') and self.ui.tray.is_tray:
            event.ignore()
            self.hide()
            return

        self.is_closing = True
        print("Closing...")
        print("Sending terminate signal to all...")
        self.controller.kernel.terminate()
        print("Saving ctx groups...")
        self.controller.ctx.save_all()
        print("Saving tabs...")
        self.core.tabs.save()
        print("Saving notepad...")
        self.controller.notepad.save_all()
        print("Saving calendar...")
        self.controller.calendar.save_all()
        print("Saving drawing...")
        self.controller.painter.save_all()
        print("Saving tools...")
        self.tools.on_exit()
        print("Saving layout state...")
        self.controller.layout.save()
        print("Stopping timers...")
        self.timer.stop()
        self.post_timer.stop()
        self.update_timer.stop()
        print("Saving config...")
        self.core.config.save()
        print("Saving presets...")
        self.core.presets.save_all()
        print("Exiting...")
        print("")
        print("Have a nice day! :)")
        event.accept()

    def changeEvent(self, event):
        """
        Handle window state change event

        :param event: Event
        """
        if event.type() == QEvent.WindowStateChange:
            if self.isMinimized() and self.core.config.get('layout.tray.minimize'):
                self.ui.tray_menu['restore'].setVisible(True)
                self.hide()
                event.ignore()
            else:
                self.prevState = event.oldState()

    def tray_toggle(self):
        """Toggle tray icon"""
        if self.core.config.get('layout.tray.minimize'):
            if self.isVisible():
                if not self.isActiveWindow():
                    self.activateWindow()
                else:
                    self.ui.tray_menu['restore'].setVisible(True)
                    self.hide()
            else:
                self.restore()
        else:
            if self.isMinimized():
                self.restore()
            else:
                if not self.isActiveWindow():
                    self.activateWindow()
                else:
                    self.showMinimized()

    def restore(self):
        """Restore window"""
        if self.prevState == Qt.WindowMaximized or self.windowState() == Qt.WindowMaximized:
            self.showMaximized()
        else:
            self.showNormal()
        self.activateWindow()
        self.ui.tray_menu['restore'].setVisible(False)

    def setup_global_shortcuts(self):
        """Setup global shortcuts"""
        if not hasattr(self, 'core') or not hasattr(self.core, 'config'):
            return

        # unregister existing shortcuts
        if hasattr(self, 'shortcuts'):
            for shortcut in self.shortcuts:
                # disconnect signals
                shortcut.activated.disconnect()
                # disable the shortcut to prevent it from handling events
                shortcut.setEnabled(False)
                # remove the parent to break the QObject tree
                shortcut.setParent(None)
                # schedule the shortcut for deletion
                shortcut.deleteLater()
            # clear the list of shortcuts
            self.shortcuts.clear()
            # process events to delete shortcuts immediately
            QtWidgets.QApplication.processEvents()
        else:
            self.shortcuts = []

        # Handle the Escape key
        escape_shortcut = QShortcut(QKeySequence(Qt.Key_Escape), self)
        escape_shortcut.setContext(Qt.ApplicationShortcut)
        escape_shortcut.activated.connect(self.controller.access.on_escape)
        self.shortcuts.append(escape_shortcut)

        config = copy.deepcopy(self.core.config.get("access.shortcuts"))
        if config is None:
            return
            
        for shortcut_conf in config:
            key = shortcut_conf.get('key', '')
            key_modifier = shortcut_conf.get('key_modifier', '')
            action_name = shortcut_conf.get('action')

            if not key or not action_name:
                continue

            key_sequence_parts = []
            if key_modifier and key_modifier != '---':
                if key_modifier == "Control":
                    key_modifier = "Ctrl"
                key_sequence_parts.append(key_modifier)
            key_sequence_parts.append(key)
            key_sequence_str = '+'.join(key_sequence_parts)
            key_sequence = QKeySequence(key_sequence_str)

            shortcut = QShortcut(key_sequence, self)
            shortcut.setContext(Qt.ApplicationShortcut)
            shortcut.activated.connect(
                lambda checked=False, action=action_name: self.dispatch(ControlEvent(action))
            )
            self.shortcuts.append(shortcut)



#!/usr/bin/env python3
# -*- coding: utf-8 -*-
# ================================================== #
# This file is a part of PYGPT package               #
# Website: https://pygpt.net                         #
# GitHub:  https://github.com/szczyglis-dev/py-gpt   #
# MIT License                                        #
# Created By  : Marcin Szczygliński                  #
# Updated Date: 2024.11.17 03:00:00                  #
# ================================================== #

from pygpt_net.ui.widget.lists.base_list_combo import BaseListCombo


class ModeCombo(BaseListCombo):
    def __init__(self, window=None, id=None):
        """
        Mode select menu

        :param window: main window
        :param id: input id
        """
        super(ModeCombo, self).__init__(window, id)

    def on_combo_change(self, index):
        """
        On combo change

        :param index: combo index
        """
        if not self.initialized or self.locked:
            return
        self.current_id = self.combo.itemData(index)
        self.window.controller.mode.select(self.current_id)



#!/usr/bin/env python3
# -*- coding: utf-8 -*-
# ================================================== #
# This file is a part of PYGPT package               #
# Website: https://pygpt.net                         #
# GitHub:  https://github.com/szczyglis-dev/py-gpt   #
# MIT License                                        #
# Created By  : Marcin Szczygliński                  #
# Updated Date: 2025.06.30 02:00:00                  #
# ================================================== #

import copy
import os
from typing import List, Dict, Optional

from pygpt_net.core.events import Event
from pygpt_net.utils import trans


class Importer:
    def __init__(self, window=None):
        """
        Models importer controller

        :param window: Window instance
        """
        self.window = window
        self.dialog = False
        self.initialized = False
        self.width = 800
        self.height = 500
        self.items_available = {}  # available models for import
        self.items_current = {}  # current models in use
        self.pending = {}  # waiting to be imported models
        self.removed = {}  # waiting to be removed models
        self.selected_available = None  # selected available model
        self.selected_current = None  # selected current model
        self.all = False  # show all models, not only available for import

    def in_current(self, model: str) -> bool:
        """
        Check if model is in current list

        :param model: model ID
        :return: True if model is in current list, False otherwise
        """
        if model in self.items_current:
            return True
        for key in list(self.items_current.keys()):
            m = self.items_current[key]
            if m.id == model:  # also check in IDs
                return True
        return False

    def change_available(self):
        """On change available model selection"""
        val = self.window.ui.nodes["models.importer.available"].selectionModel().currentIndex()
        idx = val.row()
        if idx < 0:
            self.selected_available = None
            self.window.ui.nodes["models.importer.add"].setEnabled(False)
        else:
            self.selected_available = self.get_by_idx(idx, self.items_available)
            if self.items_available.get(self.selected_available) is None:
                self.selected_available = None
                self.window.ui.nodes["models.importer.add"].setEnabled(False)
            else:
                # if not in current then enable add button
                if not self.in_current(self.selected_available):
                    self.window.ui.nodes["models.importer.add"].setEnabled(True)
                else:
                    self.window.ui.nodes["models.importer.add"].setEnabled(False)

    def change_current(self):
        """On change current model selection"""
        val = self.window.ui.nodes["models.importer.current"].selectionModel().currentIndex()
        idx = val.row()
        if idx < 0:
            self.selected_current = None
            self.window.ui.nodes["models.importer.remove"].setEnabled(False)
        else:
            self.selected_current = self.get_by_idx(idx, self.items_current)
            if self.items_current.get(self.selected_current) is None:
                self.selected_current = None
                self.window.ui.nodes["models.importer.remove"].setEnabled(False)
            else:
                if self.selected_current in self.items_current and self.items_current[self.selected_current].imported:
                    self.window.ui.nodes["models.importer.remove"].setEnabled(True)
                else:
                    self.window.ui.nodes["models.importer.remove"].setEnabled(False)

    def add(self):
        """Add model to current list"""
        if self.selected_available is None:
            self.set_status(trans('models.importer.error.add.no_model'))
            return
        if self.in_current(self.selected_available):
            self.set_status(trans('models.importer.error.add.not_exists'))
            return
        model = self.items_available[self.selected_available]
        self.items_current[self.selected_available] = model
        if self.selected_available not in self.pending:
            self.pending[self.selected_available] = model
        if self.selected_available in self.removed:
            del self.removed[self.selected_available]
        if not self.all:
            del self.items_available[self.selected_available]
        self.refresh()

    def remove(self):
        """Remove model from current list"""
        if self.selected_current is None:
            self.set_status(trans('models.importer.error.remove.no_model'))
            return
        if not self.in_current(self.selected_current):
            self.set_status(trans('models.importer.error.remove.not_exists'))
            return
        model = self.items_current[self.selected_current]
        self.items_available[self.selected_current] = model
        if self.selected_current not in self.removed:
            self.removed[self.selected_current] = model
        del self.items_current[self.selected_current]
        if self.selected_current in self.pending:
            del self.pending[self.selected_current]
        self.refresh()

    def setup(self):
        """Set up importer"""
        idx = None
        self.window.model_importer.setup(idx)  # widget dialog setup

    def toggle_editor(self):
        """Toggle models importer dialog"""
        if self.dialog:
            self.close()
        else:
            self.open()

    def open(self, force: bool = False):
        """
        Open models editor dialog

        :param force: force open dialog
        """
        if not self.initialized:
            self.setup()
            self.initialized = True
        if not self.dialog or force:
            self.pending = {}
            self.removed = {}
            self.init()
            self.window.ui.dialogs.open(
                "models.importer",
                width=self.width,
                height=self.height,
            )
            self.dialog = True

    def close(self):
        """Close models importer dialog"""
        if self.dialog:
            self.window.ui.dialogs.close('models.importer')
            self.dialog = False

    def cancel(self):
        """Cancel models importer dialog"""
        self.close()

    def init(self):
        """Initialize importer"""
        if self.initialized and self.window.ui.nodes["models.importer.available.all"].isChecked():
            self.all = True

        base_url = "http://localhost:11434"
        if 'OLLAMA_API_BASE' in os.environ:
            base_url = os.environ['OLLAMA_API_BASE']
        self.window.ui.nodes["models.importer.url"].setText(base_url)
        self.items_available = self.get_ollama_available()
        self.items_current = self.get_ollama_current()
        self.refresh()

    def toggle_all(self, all: bool):
        """
        Toggle all models visibility

        :param all: show all models, not only available for import
        """
        self.all = all
        self.refresh(reload=True)

    def set_status(self, status: str):
        """
        Set status message

        :param status: status message
        """
        if self.initialized:
            self.window.ui.nodes["models.importer.status"].setText(status)

    def get_ollama_current(self) -> Dict:
        """
        Get current ollama models

        :return: PyGPT ollama models dictionary
        """
        items = copy.deepcopy(self.window.core.models.items)
        for key in list(items.keys()):
            if (items[key].llama_index is None
                    or items[key].provider != 'ollama'):
                del items[key]
        return items

    def get_ollama_available(self) -> Dict:
        """
        Get available ollama models

        :return: Ollama API models dictionary
        """
        models = {}
        status = self.window.core.models.ollama.get_status()
        if not status['status']:
            self.set_status(trans('models.importer.error.no_connection'))
            return models
        else:
            ollama_models = status.get('models', [])
            if not ollama_models:
                self.set_status(trans('models.importer.error.no_models'))
                return models
            else:
                for model in ollama_models:
                    name = model.get('name').replace(":latest", "")
                    m = self.window.core.models.create_empty(append=False)
                    m.id = name
                    m.name = name
                    m.mode = [
                        "chat",
                        "llama_index",
                        "agent",
                        "agent_llama",
                        "expert",
                    ]
                    m.provider = 'ollama'
                    m.input = ["text"]
                    m.output = ["text"]
                    # m.llama_index['provider'] = 'ollama'
                    # m.llama_index['mode'] = ['chat']
                    m.llama_index['args'] = [
                        {
                            'name': 'model',
                            'value': name,
                            'type': 'str'
                        }
                    ]
                    """
                    m.langchain['provider'] = 'ollama'
                    m.langchain['mode'] = ['chat']
                    m.langchain['args'] = [
                        {
                            'name': 'model',
                            'value': name,
                            'type': 'str'
                        }
                    ]
                    """
                    m.imported = True
                    m.ctx = 32000  # default
                    key = m.id
                    # if key in self.items_current:
                        # key += "_imported"
                    models[key] = m
                self.set_status(trans('models.importer.loaded'))
        return models

    def from_pending(self):
        """Move pending models to base list"""
        added = False
        base_models = self.window.core.models.items
        for key in list(self.pending.keys()):
            if key not in base_models:
                base_models[key] = copy.deepcopy(self.pending[key])
                base_models[key].imported = True
                added = True
        for key in list(self.removed.keys()):
            if key in base_models:
                del base_models[key]
                added = True
        self.pending = {}
        self.removed = {}
        if added:
            self.window.core.models.save()
            self.set_status(trans('models.importer.status.imported'))

    def save(self, persist: bool = True):
        """
        Save models

        :param persist: persist to file and close dialog
        """
        self.from_pending()
        self.window.controller.model.init_list()
        self.window.controller.model.update()
        self.close()

        # dispatch on update event
        event = Event(Event.MODELS_CHANGED)
        self.window.dispatch(event, all=True)

    def refresh(self, reload: bool = False):
        """
        Reload items

        :param reload: reload available models
        """
        if reload:
            self.items_available = self.get_ollama_available()

        # remove from available if already in current
        if not self.all:
            for key in list(self.items_available.keys()):
                if self.in_current(key):
                    del self.items_available[key]

        self.window.ui.nodes['models.importer.editor'].update_available(self.items_available)
        self.window.ui.nodes['models.importer.editor'].update_current(self.items_current)

    def get_by_idx(self, idx: int, items: Dict) -> Optional[str]:
        """
        Get model key by list index

        :param idx: list index
        :param items: items dictionary
        :return: model key
        """
        model_idx = 0
        for id in self.get_ids(items):
            if model_idx == idx:
                return id
            model_idx += 1
        return None

    def get_ids(self, items: Dict) -> List[str]:
        """
        Return models ids

        :return: model ids list
        """
        return list(items.keys())


#!/usr/bin/env python3
# -*- coding: utf-8 -*-
# ================================================== #
# This file is a part of PYGPT package               #
# Website: https://pygpt.net                         #
# GitHub:  https://github.com/szczyglis-dev/py-gpt   #
# MIT License                                        #
# Created By  : Marcin Szczygliński                  #
# Updated Date: 2025.07.23 01:00:00                  #
# ================================================== #

import json
import os
import re
from typing import List, Dict, Any

from llama_index.core.chat_engine.types import AgentChatResponse
from llama_index.core.tools import BaseTool, FunctionTool, QueryEngineTool, ToolMetadata

from pygpt_net.core.bridge.context import BridgeContext
from pygpt_net.core.events import Event
from pygpt_net.item.ctx import CtxItem


class Tools:
    def __init__(self, window=None):
        """
        Agent tools

        :param window: Window instance
        """
        self.window = window
        self.cmd_blacklist = []
        self.verbose = False
        self.code_execute_fn = CodeExecutor(window)
        self.last_tool_output = None
        self.agent_idx = None  # agent index, used for query engine tool
        self.context = None  # BridgeContext instance, used for tool execution

    def prepare(
            self,
            context: BridgeContext,
            extra: Dict[str, Any],
            verbose: bool = False,
            force: bool = False
    ) -> List[BaseTool]:
        """
        Prepare tools for agent

        :param context: BridgeContext
        :param extra: extra data
        :param verbose: verbose mode
        :param force: force to get functions even if not needed
        :return: list of tools
        """
        self.verbose = verbose
        tools = []

        # add functions from plugins
        plugin_functions = self.get_plugin_functions(context.ctx, verbose=verbose, force=force)
        tools.extend(plugin_functions)

        # add query engine tool if idx is provided
        idx = extra.get("agent_idx", None)
        if idx is not None and idx != "_":
            llm, embed_model = self.window.core.idx.llm.get_service_context(model=context.model)
            index = self.window.core.idx.storage.get(idx, llm, embed_model)  # get index
            if index is not None:
                query_engine = index.as_query_engine(similarity_top_k=3)
                query_engine_tools = [
                    QueryEngineTool(
                        query_engine=query_engine,
                        metadata=ToolMetadata(
                            name="query_engine",
                            description=(
                                "Provides additional context and access to the indexed documents."
                            ),
                        ),
                    ),
                ]
                tools.extend(query_engine_tools)
        return tools

    def get_plugin_functions(
            self,
            ctx: CtxItem,
            verbose: bool = False,
            force: bool = False
    ) -> list:
        """
        Parse plugin functions

        :param ctx: CtxItem
        :param verbose: verbose mode
        :param force: force to get functions even if not needed
        :return: List of functions
        """
        tools = []
        functions = self.window.core.command.get_functions(force=force)
        for func in functions:
            try:
                name = func['name']
                if name in self.cmd_blacklist:
                    continue  # skip blacklisted commands

                description = func['desc']
                schema = json.loads(func['params'])  # from JSON to dict

                def make_func(name):
                    def func(**kwargs):
                        self.log("[Plugin] Tool call: " + name + " " + str(kwargs))
                        cmd = {
                            "cmd": name,
                            "params": kwargs,
                        }
                        response = self.window.controller.plugins.apply_cmds_all(
                            ctx,  # current ctx
                            [cmd],  # commands
                        )
                        return str(response)  # return response as string

                    return func

                func = make_func(name)
                metadata = PluginToolMetadata(
                    name=name,
                    description=description,
                )
                metadata.schema = schema
                tool = FunctionTool(
                    fn=func,
                    metadata=metadata,
                )
                tools.append(tool)
            except Exception as e:
                print(e)
        return tools

    def get_plugin_tools(
            self,
            context: BridgeContext,
            extra: Dict[str, Any],
            verbose: bool = False
    ) -> dict:
        """
        Parse plugin functions

        :param context: BridgeContext
        :param extra: extra data
        :param verbose: verbose mode
        """
        tools = {}
        functions = self.window.core.command.get_functions()
        for func in functions:
            try:
                name = func['name']
                if name in self.cmd_blacklist:
                    continue  # skip blacklisted commands

                description = func['desc']
                schema = json.loads(func['params'])  # from JSON to dict

                def make_func(name):
                    def func(**kwargs):
                        self.log("[Plugin] Tool call: " + name + " " + str(kwargs))
                        cmd = {
                            "cmd": name,
                            "params": kwargs,
                        }
                        response = self.window.controller.plugins.apply_cmds_all(
                            context.ctx,  # current ctx
                            [cmd],  # commands
                        )
                        return str(response)  # return response as string

                    return func

                func = make_func(name)
                tools[name] = func
            except Exception as e:
                print(e)

        # add query engine tool if idx is provided
        if self.agent_idx is not None and self.agent_idx  != "_":
            tools["query_engine"] = None  # placeholder for query engine tool

        return tools


    def get_plugin_specs(
            self,
            context: BridgeContext,
            extra: Dict[str, Any],
            verbose: bool = False
    ) -> list:
        """
        Parse plugin functions

        :param context: BridgeContext
        :param extra: extra data
        :param verbose: verbose mode
        """
        specs = []
        functions = self.window.core.command.get_functions()

        # add query engine tool spec if idx is provided
        if self.agent_idx is not None and self.agent_idx  != "_":
            specs.append("**query_engine**: "
                         "Provides additional context and access to the indexed documents, "
                         "available params: {'query': {'type': 'string', 'description': 'query string'}}, required: [query]")

        for func in functions:
            try:
                name = func['name']
                if name in self.cmd_blacklist:
                    continue  # skip blacklisted commands
                description = func['desc']
                schema = json.loads(func['params'])  # from JSON to dict
                spec = "**{}**: {}, available params: {}, required: {}\n".format(name, description, schema.get("properties", {}), schema.get("required", []))
                specs.append(spec)
            except Exception as e:
                print(e)
        return specs

    def tool_exec(self, cmd: str, params: Dict[str, Any]) -> str:
        """
        Execute tool command

        :param cmd: command name
        :param params: command parameters
        :return: command output
        """
        print("[Plugin] Tool call: " + cmd + " " + str(params))
        # special case for query engine tool
        if cmd == "query_engine":
            if "query" not in params:
                return "Query parameter is required for query_engine tool."
            if self.context is None:
                return "Context is not set for query_engine tool."
            if self.agent_idx is None or self.agent_idx == "_":
                return "Agent index is not set for query_engine tool."
            llm, embed_model = self.window.core.idx.llm.get_service_context(model=self.context.model)
            index = self.window.core.idx.storage.get(self.agent_idx, llm, embed_model)  # get index
            if index is not None:
                query_engine = index.as_query_engine(similarity_top_k=3)
                response = query_engine.query(params["query"])
                print("[Plugin] Query engine response: " + str(response))
                self.log("[Plugin] Query engine response: " + str(response))
                return str(response)
            else:
                return "Index not found for query_engine tool."

        # rest of the plugin commands
        cmd = {
            "cmd": cmd,
            "params": params,
        }
        ctx = CtxItem()
        ctx.extra["agent_input"] = True  # mark as user input
        ctx.agent_call = True  # disables reply from plugin commands
        response = self.window.controller.plugins.apply_cmds_all(
            ctx,  # current ctx
            [cmd],  # commands
        )
        return response

    def get_retriever_tool(
            self,
            context: BridgeContext,
            extra: Dict[str, Any],
            verbose: bool = False
    ) -> List[BaseTool]:
        """
        Prepare tools for agent

        :param context: BridgeContext
        :param extra: extra data
        :param verbose: verbose mode
        :return: list of tools
        """
        tool = None
        # add query engine tool if idx is provided
        idx = extra.get("agent_idx", None)
        if idx is not None and idx != "_":
            llm, embed_model = self.window.core.idx.llm.get_service_context(model=context.model)
            index = self.window.core.idx.storage.get(idx, llm, embed_model)  # get index
            if index is not None:
                query_engine = index.as_query_engine(similarity_top_k=3)
                tool = [
                    QueryEngineTool(
                        query_engine=query_engine,
                        metadata=ToolMetadata(
                            name="query_engine",
                            description=(
                                "Provides additional context and access to the indexed documents."
                            ),
                        ),
                    ),
                ]
        return tool

    def export_sources(
            self,
            response: AgentChatResponse
    ) -> List[dict]:
        """
        Export sources from response

        :param response: response
        :return: list of sources
        """
        data = []
        for output in response.sources:
            item = {}
            item["ToolOutput"] = {
                "content": str(output.content),
                "tool_name": str(output.tool_name),
                "raw_input": str(output.raw_input),
                "raw_output": str(output.raw_output),
            }
            data.append(item)
        return data

    def get_last_tool_output(self) -> dict:
        """
        Get last tool output

        :return: last tool output
        """
        if self.window.core.agents.tools.last_tool_output is None:
            return {}
        return self.window.core.agents.tools.last_tool_output

    def has_last_tool_output(self) -> bool:
        """
        Check if there is a last tool output

        :return: True if last tool output exists, False otherwise
        """
        return self.window.core.agents.tools.last_tool_output is not None

    def clear_last_tool_output(self):
        """Clear last tool output"""
        self.window.core.agents.tools.last_tool_output = None

    def append_tool_outputs(self, ctx: CtxItem, clear: bool = True):
        """
        Append tool outputs to context

        :param ctx: CtxItem
        :param clear: clear last tool output after appending
        """
        if self.has_last_tool_output():
            outputs = [self.get_last_tool_output()]
            ctx.extra["tool_output"] = outputs
            if outputs is not None:
                response = ""
                for output in outputs:
                    if ("code" in output and "output" in output["code"] and
                            "content" in output["code"]["output"]):
                        response += str(output["code"]["output"]["content"])
                self.window.core.filesystem.parser.extract_data_files(ctx, response) # img, files
            if clear:
                self.clear_last_tool_output()  # clear after use

    def extract_tool_outputs(self, ctx: CtxItem, clear: bool = True):
        """
        Append tool outputs to context

        :param ctx: CtxItem
        :param clear: clear last tool output after appending
        """
        if self.has_last_tool_output():
            outputs = [self.get_last_tool_output()]
            if outputs is not None:
                response = ""
                for output in outputs:
                    if ("code" in output and "output" in output["code"] and
                            "content" in output["code"]["output"]):
                        response += str(output["code"]["output"]["content"])
                self.window.core.filesystem.parser.extract_data_files(ctx, response) # img, files
            if clear:
                self.clear_last_tool_output()  # clear after use

    def log(self, msg: str):
        """
        Log message

        :param msg: message
        """
        if self.verbose:
            print(msg)
            self.window.core.debug.add(msg)

class PluginToolMetadata(ToolMetadata):
    def __init__(self, name: str, description: str):
        super().__init__(name=name, description=description)
        self.schema = None

    def get_parameters_dict(self) -> Dict[str, Any]:
        """
        Get parameters dictionary

        :return: parameters
        """
        parameters = {
            k: v
            for k, v in self.schema.items()
            if k in ["type", "properties", "required", "definitions"]
        }
        return parameters

class CodeExecutor:
    """Code executor for codeAct agent"""

    def __init__(self, window = None):
        """
        Initialize the code executor.

        :param window: Window instance
        """
        self.window = window

    def execute(self, code: str) -> str:
        """
        Execute Python code and capture output and return values.

        :param code: Python code to execute
        :return: Output from the code execution
        """
        self.window.core.agents.tools.last_tool_output = None
        if code == "/restart":
            commands = [
                {
                    "cmd": "ipython_kernel_restart",
                    "params": {},
                    "silent": True,
                    "force": True,
                }
            ]
        else:
            commands = [
                {
                    "cmd": "ipython_execute",
                    "params": {
                        "code": code,
                        "path": ".interpreter.current.py",
                    },
                    "silent": True,
                    "force": True,
                }
            ]
        event = Event(Event.CMD_EXECUTE, {
            'commands': commands,
            'silent': True,
        })
        event.ctx = CtxItem()  # tmp
        event.ctx.async_disabled = True  # disable async for this event
        self.window.controller.command.dispatch_only(event)

        # if restart command was executed, return success message
        if code == "/restart":
            return "IPython kernel restarted successfully."

        response = event.ctx.bag  # tmp response
        output = ""

        # store tool output if available
        if "code" in response:
            if "output" in response["code"]:
                output = response["code"]["output"]["content"]
                tool_output = {
                    "cmd": "ipython_execute",
                    "code": {
                        "input": {
                            "content": response["code"]["input"]["content"],
                            "lang": "python"
                        },
                        "output": {
                            "content": response["code"]["output"]["content"],
                            "lang": "python"
                        }
                    },
                    "plugin": response["plugin"],
                    "result": response["result"]
                }
                self.window.core.agents.tools.last_tool_output = tool_output
        return output



#!/usr/bin/env python3
# -*- coding: utf-8 -*-
# ================================================== #
# This file is a part of PYGPT package               #
# Website: https://pygpt.net                         #
# GitHub:  https://github.com/szczyglis-dev/py-gpt   #
# MIT License                                        #
# Created By  : Marcin Szczygliński                  #
# Updated Date: 2025.07.23 01:00:00                  #
# ================================================== #

import json
from typing import Dict, List, Optional

from PySide6.QtCore import QRunnable, QObject, Signal, Slot
from llama_index.core.base.llms.types import ChatMessage, MessageRole
from llama_index.core.tools import QueryEngineTool

from pygpt_net.core.types import (
    MODE_AGENT,
    MODE_CHAT,
    MODE_COMPLETION,
    MODE_EXPERT,
    MODE_LANGCHAIN,
    MODE_LLAMA_INDEX,
    MODE_VISION,
    MODE_AUDIO,
    MODE_RESEARCH,
)
from pygpt_net.core.bridge.context import BridgeContext
from pygpt_net.core.events import Event, KernelEvent, RenderEvent
from pygpt_net.item.ctx import CtxItem
from pygpt_net.item.preset import PresetItem
from pygpt_net.utils import trans


class Experts:
    def __init__(self, window=None):
        """
        Experts core

        :param window: Window instance
        """
        self.window = window
        self.allowed_modes = [
            MODE_CHAT,
            MODE_COMPLETION,
            MODE_VISION,
            # MODE_LANGCHAIN,
            MODE_LLAMA_INDEX,
            MODE_AUDIO,
            MODE_RESEARCH,
        ]
        self.allowed_cmds = ["expert_call"]
        self.worker = None

    def get_mode(self) -> str:
        """
        Get sub-mode to use internally

        :return: sub-mode
        """
        mode = MODE_CHAT
        current = self.window.core.config.get("experts.mode")
        if current is not None and current in self.allowed_modes:
            mode = current
        return mode

    def stopped(self) -> bool:
        """
        Check if experts are stopped

        :return: True if stopped
        """
        return self.window.controller.kernel.stopped()

    def agent_enabled(self) -> bool:
        """
        Check if agent is enabled

        :return: True if enabled
        """
        return self.window.controller.agent.legacy.enabled()

    def exists(self, id: str) -> bool:
        """
        Check if expert exists

        :param id: expert id
        :return: True if exists
        """
        return self.window.core.presets.has(MODE_EXPERT, id)

    def get_expert(self, id: str) -> PresetItem:
        """
        Get expert by id

        :param id: expert id
        :return: expert item (preset)
        """
        return self.window.core.presets.get_by_id(MODE_EXPERT, id)

    def get_experts(self) -> Dict[str, str]:
        """
        Get experts names with keys

        :return: experts dict
        """
        experts = {}
        presets = self.window.core.presets.get_by_mode(MODE_EXPERT)

        # mode: agent
        if self.agent_enabled():
            agents = self.window.core.presets.get_by_mode(MODE_AGENT)
            agent = self.window.core.config.get('preset')
            if agent is not None:
                if agent in agents:
                    for uuid in agents[agent].experts:
                        expert = self.window.core.presets.get_by_uuid(uuid)
                        if expert is not None:
                            id = expert.filename
                            experts[id] = expert.name
        # mode: expert
        else:
            for k in presets:
                if k.startswith("current."):  # skip current
                    continue
                if not presets[k].enabled:  # skip disabled experts
                    continue
                experts[k] = presets[k].name
        return experts

    def get_expert_name_by_id(self, id: str) -> str:
        """
        Get expert name by id

        :param id: expert id
        :return: expert name
        """
        experts = self.get_experts()
        if id in experts:
            return experts[id]

    def count_experts(self, uuid: str) -> int:
        """
        Count experts in agent

        :param uuid: agent uuid
        :return: number of experts
        """
        i = 0
        agents = self.window.core.presets.get_by_mode(MODE_AGENT)
        if uuid in agents:
            for expert_uuid in agents[uuid].experts:
                expert = self.window.core.presets.get_by_uuid(expert_uuid)
                if expert is not None:
                    i += 1
        return i

    def get_prompt(self) -> str:
        """
        Get prompt for handle experts

        :return: prompt with list of experts appended
        """
        prompt = self.window.core.config.get('prompt.expert')
        if self.window.core.command.is_native_enabled():
            prompt = """
            EXPERT MODE IS ENABLED
            ======================
            You are the professional manager of a team of experts, each of whom is the best in their respective fields. 

            IMPORTANT RULES:

            YOU MUST ALWAYS FOLLOW THIS RULES WHILE MANAGING EXPERTS:

            - Instead of directly answering questions, you can manage these experts by forwarding queries to them, 
            collecting their responses, and controlling the flow of information between them as well as sending 
            requests to them. 
            - You will serve as a bridge between them. 
            - Each expert has their own conversation context, so you can communicate with them normally. 
            - The experts can communicate with you but not with each other, so you will be responsible for relaying 
            information between them. 
            - Each expert can use tools and commands, for e.g. reading files or search the Web. 
            - You can ask experts for calling tools and commands.
            - After consulting with the experts, your task will be to prepare the best possible answer based on the 
            assistance received and to use it in your Chain-of-Thought. 
            - To send a request to an expert prepare the "expert_call" tool call.
            - Remember that the call must contain a query and the ID of the expert to be forwarded to the expert. 
            - Use ONLY experts listed below, DO NOT TRY to ask any non-existing experts - they will not respond.
            - Do not respond to the expert's replies, or use commands again if replying; the expert only has 
            the ability to listen to your queries sent and to respond to them.
            - Always follow ALL user commands, even if you thing than expert can't do something. 
            Experts are allowed to execute commands same as you can.
            - After query an expert, always wait for expert response before sending another query.
            - Do not try to query experts in the middle of the conversation, put your queries and commands at the end of the text.
            - Always call only one expert at a time.
            - Always execute only one tool or function call at a time.
            - Always use the language used by user in conversation.
            - Never write as an called expert and do not imagine response from him, instead always wait for response from expert.
            - Your response cannot include <tool>...</tool> commands in the conversation body - this syntax is reserved only for response purposes.
            - Every message sent to you directly after your query is always a response from asked expert, not from me.
            - Always send a direct, clear and detailed prompt to the expert, without any "Please...", "Could you..", etc.
            - Always send to expert all the messages that user wants to send to expert, without any exceptions.

            List of available experts, along with their IDs:

            {presets}
            """
        experts = self.get_experts()
        experts_list = []
        for k in experts:
            if k.startswith("current."):  # skip current presets
                continue
            experts_list.append(" - " + str(k) + ": " + str(experts[k]))
        prompt = prompt.replace("{presets}", "\n".join(experts_list))
        return prompt

    def extract_calls(self, ctx: CtxItem) -> Dict[str, str]:
        """
        Extract expert calls from context output

        :param ctx: context item
        :return: dict with calls
        """
        ids = self.get_experts().keys()
        if not ids:  # abort if no experts
            return {}
        cmds = self.window.core.command.extract_cmds(ctx.output)
        if len(cmds) > 0:
            ctx.cmds = cmds  # append commands to ctx
        else:  # abort if no cmds
            return {}
        commands = self.window.core.command.from_commands(cmds)  # pack to execution list
        is_cmd = False
        my_commands = []
        calls = {}
        for item in commands:
            if item["cmd"] in self.allowed_cmds:
                my_commands.append(item)
                is_cmd = True
        if not is_cmd:  # abort if no expert calls
            return {}
        for item in my_commands:
            try:
                if item["cmd"] == "expert_call":
                    if "params" not in item:
                        continue
                    if "id" not in item["params"] or "query" not in item["params"]:
                        continue
                    id = item["params"]["id"]
                    if id not in ids:
                        continue
                    query = item["params"]["query"]
                    calls[id] = query
            except Exception as e:
                self.window.core.debug.error(e)
                return {}
        return calls

    def reply(self, ctx: CtxItem):
        """
        Re-send response from commands to master expert

        If command has been called by expert then response for command is send here to parent

        :param ctx: context item
        """
        if self.stopped():
            return

        # master meta is here, ctx.meta = MASTER META ID here!

        # make copy of ctx for reply, and change input name to expert name
        reply_ctx = CtxItem()
        reply_ctx.from_dict(ctx.to_dict())

        # reply ctx has no meta here!!!!
        reply_ctx.input_name = "Expert"
        reply_ctx.output_name = ""
        reply_ctx.sub_call = True  # this flag is not copied in to_dict

        internal = False
        if self.agent_enabled():  # hide in agent mode
            internal = True
        if ctx.output.strip() != "":
            response = reply_ctx.output
        else:
            response = reply_ctx.input

        context = BridgeContext()
        context.ctx = reply_ctx
        context.prompt = "Result from expert:\n\n" + str(response)
        extra = {
            "force": True,
            "reply": True,
            "internal": internal,
        }
        event = KernelEvent(KernelEvent.INPUT_SYSTEM, {
            'context': context,
            'extra': extra,
        })
        self.window.dispatch(event)

    def call(
            self,
            master_ctx: CtxItem,
            expert_id: str,
            query: str
    ):
        """
        Call the expert

        :param master_ctx: master context
        :param expert_id: expert id (preset ID)
        :param query: input text (master prompt)
        """
        if self.stopped():
            return

        self.worker = ExpertWorker(
            window=self.window,
            master_ctx=master_ctx,
            expert_id=expert_id,
            query=query
        )
        self.worker.signals.response.connect(self.handle_response)  # connect to finished signal
        self.worker.signals.finished.connect(self.handle_finished)  # connect to finished signal
        self.worker.signals.error.connect(self.handle_error)  # connect to error signal
        self.worker.signals.event.connect(self.handle_event)  # connect to event signal
        self.worker.signals.output.connect(self.handle_output)  # connect to output signal
        self.worker.signals.lock_input.connect(self.handle_input_locked)  # connect to lock input signal
        self.worker.signals.cmd.connect(self.handle_cmd)  # connect to cmd signal

        # start worker in thread pool
        expert_name = self.get_expert_name_by_id(expert_id)
        event = KernelEvent(KernelEvent.STATE_BUSY, {
            "msg": trans("expert.wait.status") + " ({})".format(expert_name),
        })
        self.window.dispatch(event)  # dispatch busy event
        self.window.threadpool.start(self.worker)

    @Slot(CtxItem, str)
    def handle_output(self, ctx: CtxItem, mode: str):
        """
        Handle output from worker

        :param ctx: CtxItem
        :param mode: mode of operation
        """
        if self.stopped():
            return

        self.window.controller.chat.output.handle(
            ctx=ctx,
            mode=mode,
            stream_mode=False,
        )

    @Slot(CtxItem)
    def handle_cmd(self, ctx: CtxItem):
        """
        Handle command from worker

        :param ctx: CtxItem
        """
        if self.stopped():
            return
        self.window.controller.chat.command.handle(ctx)  # handle cmds
        self.window.controller.kernel.stack.handle()  # handle command queue

    @Slot()
    def handle_input_locked(self):
        """
        Handle input locked signal from worker

        This is used to lock input in UI while worker is processing.
        """
        if self.stopped():
            return
        self.window.controller.chat.input.lock()

    @Slot(Event)
    def handle_event(self, event: Event):
        """
        Handle event from worker

        :param event: Event instance
        """
        if self.stopped():
            return
        self.window.dispatch(event)

    @Slot(str)
    def handle_error(self, error: str):
        """
        Handle worker error

        :param error: error message
        """
        if self.stopped():
            event = KernelEvent(KernelEvent.STATE_IDLE, {})
            self.window.dispatch(event)  # dispatch idle event
            return

        # handle error from worker
        context = BridgeContext()
        context.prompt = trans("expert.wait.failed") + ": " + str(error)
        extra = {
            "force": True,
            "reply": False,
            "internal": False,
        }
        # reply to master
        event = KernelEvent(KernelEvent.INPUT_SYSTEM, {
            'context': context,
            'extra': extra,
        })
        self.window.dispatch(event)
        event = KernelEvent(KernelEvent.STATE_IDLE, {})
        self.window.dispatch(event)  # dispatch idle event

    @Slot(CtxItem, str)
    def handle_finished(self):
        """Handle worker finished signal"""
        event = KernelEvent(KernelEvent.STATE_IDLE, {})
        self.window.dispatch(event)  # dispatch idle event


    @Slot(CtxItem, str)
    def handle_response(self, ctx: CtxItem, expert_id: str):
        """
        Handle response expert call

        :param ctx: CtxItem
        :param expert_id: expert id
        """
        if self.stopped():
            event = KernelEvent(KernelEvent.STATE_IDLE, {})
            self.window.dispatch(event)  # dispatch idle event
            return

        # handle reply from worker
        context = BridgeContext()
        context.ctx = ctx
        result = {
            "expert_id": expert_id,
            "result": str(ctx.output),
        }
        context.prompt = json.dumps(result, ensure_ascii=False, indent=2)  # prepare prompt for reply
        extra = {
            "force": True,
            "reply": False,
            "internal": False,
        }
        # reply to master
        event = KernelEvent(KernelEvent.INPUT_SYSTEM, {
            'context': context,
            'extra': extra,
        })
        self.window.dispatch(event)
        event = KernelEvent(KernelEvent.STATE_IDLE, {})
        self.window.dispatch(event)  # dispatch idle event

    def get_functions(self) -> List[Dict[str, str]]:
        """
        Append call the expert commands

        :return: call the expert commands
        """
        cmds = [
            {
                "cmd": "expert_call",
                "instruction": "Call the expert",
                "params": [
                    {
                        "name": "id",
                        "description": "expert id",
                        "required": True,
                        "type": "str",
                    },
                    {
                        "name": "query",
                        "description": "query to expert",
                        "required": True,
                        "type": "str",
                    }
                ]
            }
        ]
        return cmds

    def has_calls(self, ctx: CtxItem) -> bool:
        """
        Check if context has expert calls

        :param ctx: CtxItem
        """
        if not ctx.sub_reply and not ctx.reply:
            mentions = self.window.core.experts.extract_calls(ctx)
            if mentions:
                for expert_id in mentions:
                    if not self.window.core.experts.exists(expert_id):
                        continue
                    return True
        return False


class WorkerSignals(QObject):
    """
    Signals for worker to communicate with main thread.
    """
    finished = Signal()  # when worker is finished
    response = Signal(object, str)  # when worker has response
    error = Signal(str)  # when worker has error
    event = Signal(object)  # when worker has event to dispatch
    output = Signal(object, str)  # when worker has output to handle
    lock_input = Signal()  # when worker locks input for UI
    cmd = Signal(object)  # when worker has command to execute


class ExpertWorker(QObject, QRunnable):
    """Worker for handling expert calls in a separate thread."""
    def __init__(
            self,
            window,
            master_ctx: CtxItem,
            expert_id: str,
            query: str
    ):
        QObject.__init__(self)
        QRunnable.__init__(self)
        self.window = window
        self.master_ctx = master_ctx
        self.expert_id = expert_id
        self.query = query
        self.signals = WorkerSignals()

    @Slot()
    def run(self):
        master_ctx = self.master_ctx
        expert_id = self.expert_id
        query = self.query

        try:
            # get or create children (slave) meta
            slave = self.window.core.ctx.get_or_create_slave_meta(master_ctx, expert_id)
            expert = self.window.core.experts.get_expert(expert_id)
            reply = True
            hidden = False
            internal = False

            if self.window.core.experts.agent_enabled():  # hide in agent mode
                internal = False
                hidden = True

            mode = self.window.core.config.get("mode")
            base_mode = mode
            model = expert.model
            expert_name = expert.name
            ai_name = ""
            sys_prompt = expert.prompt
            model_data = self.window.core.models.get(model)

            files = []
            file_ids = []
            functions = []
            tools_outputs = []

            # from current config
            max_tokens = self.window.core.config.get('max_output_tokens')
            stream_mode = self.window.core.config.get('stream')
            verbose = self.window.core.config.get('agent.llama.verbose')
            use_agent = self.window.core.config.get('experts.use_agent', False)
            db_idx = self.window.controller.idx.current_idx  # get idx from agent config

            mode = MODE_EXPERT  # force expert mode, mode will change in bridge

            # create slave item
            ctx = CtxItem()
            ctx.meta = slave  # use slave-meta
            ctx.internal = internal
            ctx.hidden = hidden
            ctx.current = True  # mark as current context item
            ctx.mode = mode  # store current selected mode (not inline changed)
            ctx.model = model  # store model list key, not real model id
            ctx.set_input(query, str(ai_name))
            ctx.set_output(None, expert_name)
            ctx.sub_call = True  # mark as sub-call
            ctx.pid = master_ctx.pid  # copy PID from parent to allow reply

            # render: begin
            event = RenderEvent(RenderEvent.BEGIN, {
                "meta": ctx.meta,
                "ctx": ctx,
                "stream": stream_mode,
            })
            self.signals.event.emit(event)  # dispatch render event
            self.window.core.ctx.provider.append_item(slave, ctx)  # to slave meta

            # build sys prompt
            sys_prompt_raw = sys_prompt  # store raw prompt
            event = Event(Event.PRE_PROMPT, {
                'mode': mode,
                'value': sys_prompt,
                'is_expert': True,
            })
            self.signals.event.emit(event)  # dispatch pre-prompt event
            sys_prompt = event.data['value']
            sys_prompt = self.window.core.prompt.prepare_sys_prompt(
                mode,
                model_data,
                sys_prompt,
                ctx,
                reply,
                internal,
                is_expert=True,  # mark as expert, blocks expert prompt append in plugin
            )

            # index to use
            use_index = False
            if db_idx:
                use_index = True
            if use_index:
                index, llm = self.window.core.idx.chat.get_index(db_idx, model_data, stream=False)
            else:
                llm = self.window.core.idx.llm.get(model_data, stream=False)

            history = self.window.core.ctx.all(
                meta_id=slave.id
            )  # get history for slave ctx, not master ctx

            if use_agent:
                # call the agent (planner) with tools and index
                ctx.agent_call = True  # directly return tool call response
                ctx.use_agent_final_response = True  # use agent final response as output
                bridge_context = BridgeContext(
                    ctx=ctx,
                    history=history,
                    mode=mode,
                    parent_mode=base_mode,
                    model=model_data,
                    system_prompt=sys_prompt,
                    system_prompt_raw=sys_prompt_raw,
                    prompt=query,
                    stream=False,
                    attachments=files,
                    file_ids=file_ids,
                    assistant_id=self.window.core.config.get('assistant'),
                    idx=db_idx,
                    idx_mode=self.window.core.config.get('llama.idx.mode'),
                    external_functions=functions,
                    tools_outputs=tools_outputs,
                    max_tokens=max_tokens,
                )
                extra = {}
                if use_index:
                    extra["agent_idx"] = db_idx
                tools = self.window.core.agents.tools.prepare(
                    bridge_context, extra, verbose=False, force=True)

                # remove expert_call tool from tools
                for tool in list(tools):
                    if tool.metadata.name == "expert_call":
                        tools.remove(tool)

                result = self.call_agent(
                    context=bridge_context,
                    tools=tools,
                    ctx=ctx,
                    query=query,
                    llm=llm,
                    system_prompt=sys_prompt,
                    verbose=verbose,
                )
                ctx.reply = False  # reset reply flag, we handle reply here
            else:
                # call bridge
                bridge_context = BridgeContext(
                    ctx=ctx,
                    history=history,
                    mode=mode,
                    parent_mode=base_mode,
                    model=model_data,
                    system_prompt=sys_prompt,
                    system_prompt_raw=sys_prompt_raw,
                    prompt=query,
                    stream=False,
                    attachments=files,
                    file_ids=file_ids,
                    assistant_id=self.window.core.config.get('assistant'),
                    idx=db_idx,
                    idx_mode=self.window.core.config.get('llama.idx.mode'),
                    external_functions=functions,
                    tools_outputs=tools_outputs,
                    max_tokens=max_tokens,
                )

                self.signals.lock_input.emit()  # emit lock input signal
                event = KernelEvent(KernelEvent.CALL, {
                    'context': bridge_context,  # call using slave ctx history
                    'extra': {},
                })
                self.window.dispatch(event)
                result = event.data.get("response")

            if not result:  # abort if bridge call failed
                self.signals.finished.emit()
                return

            # handle output
            ctx.current = False  # reset current state
            ctx.output = result  # store expert output in their context

            self.window.core.ctx.update_item(ctx)

            ctx.from_previous()  # append previous result if exists

            # self.window.core.debug.pause()  # pause debug if enabled

            # tmp switch meta for render purposes
            ctx.meta = master_ctx.meta
            self.signals.output.emit(ctx, mode)  # emit output signal

            ctx.clear_reply()  # reset results
            ctx.meta = slave  # restore before cmd execute

            if not use_agent:
                self.signals.cmd.emit(ctx)  # emit cmd signal to handle commands

            # if command to execute then end here, and reply is returned to reply() above from stack, and ctx.reply = TRUE here
            ctx.from_previous()  # append previous result again before save
            self.window.core.ctx.update_item(ctx)  # update ctx in DB

            # if commands reply after bridge call, then stop (already handled in sync dispatcher)
            if ctx.reply:
                self.signals.finished.emit()
                return

            # make copy of ctx for reply, and change input name to expert name
            reply_ctx = CtxItem()

            reply_ctx.from_dict(ctx.to_dict())
            reply_ctx.meta = master_ctx.meta

            # assign expert output
            reply_ctx.output = result
            reply_ctx.input_name = expert_name
            reply_ctx.output_name = ""
            reply_ctx.cmds = []  # clear cmds
            reply_ctx.sub_call = True  # this flag is not copied in to_dict

            # reply to main thread
            self.signals.response.emit(reply_ctx, str(expert_id))
        except Exception as e:
            self.window.core.debug.log(e)
            self.signals.error.emit(str(e))
        finally:
            self.signals.finished.emit()

    def call_agent(
            self,
            context: BridgeContext,
            tools: Optional[List[QueryEngineTool]] = None,
            ctx: Optional[CtxItem] = None,
            query: str = "",
            llm=None,
            system_prompt: str = "",
            verbose: bool = False,

    ) -> str:
        """
        Call agent with tools and index

        :param context: Bridge context
        :param tools: Tools
        :param ctx: CtxItem
        :param query: Input prompt
        :param llm: LLM provider
        :param system_prompt: System prompt to use for agent
        :param verbose: Verbose mode, default is False
        :return: True if success, False otherwise
        """
        history = self.window.core.agents.memory.prepare(context)
        if system_prompt:
            msg = ChatMessage(
                role=MessageRole.SYSTEM,
                content=system_prompt
            )
            history.insert(0, msg)
        kwargs = {
            "context": context,
            "tools": tools,
            "llm": llm,
            "chat_history": history,
            "max_iterations": 30,
            "verbose": verbose,
            "system_prompt": system_prompt,
            "are_commands": self.window.core.config.get("cmd"),
        }
        provider = self.window.core.agents.provider.get("planner")
        agent = provider.get_agent(self.window, kwargs)
        response_ctx = self.window.core.agents.runner.run_plan_once(
            agent=agent,
            ctx=ctx,
            prompt=query,
            verbose=verbose,
        )
        if response_ctx:
            return str(response_ctx.output)
        else:
            return "No response from expert."


#!/usr/bin/env python3
# -*- coding: utf-8 -*-
# ================================================== #
# This file is a part of the PYGPT package           #
# Website: https://pygpt.net                         #
# GitHub:  https://github.com/szczyglis-dev/py-gpt   #
# MIT License                                        #
# Created By  : Marcin Szczygliński                  #
# Updated Date: 2025.07.22 15:00:00                  #
# ================================================== #

from PySide6.QtGui import QAction, QIcon
from PySide6.QtWidgets import QLineEdit, QStyle
from PySide6.QtCore import QTimer

from pygpt_net.utils import trans
import pygpt_net.icons_rc


class CtxSearchInput(QLineEdit):
    def __init__(self, window=None):
        """
        Search input

        :param window: Window instance
        """
        super(CtxSearchInput, self).__init__(window)
        self.window = window
        self.setPlaceholderText(trans('ctx.list.search.placeholder'))
        self.setProperty('class', 'layout-search')

        # action for clearing the search
        self.clear_action = QAction(self)
        self.clear_action.setIcon(QIcon(":/icons/close.svg"))
        self.clear_action.triggered.connect(self.clear_search_string)
        self.addAction(self.clear_action, QLineEdit.TrailingPosition)
        self.clear_action.setVisible(False)

        # search action - icon on the left
        action = QAction(self)
        action.setIcon(QIcon(":/icons/search.svg"))
        self.addAction(action, QLineEdit.LeadingPosition)

        # timer to delay search_string_change
        self._search_timer = QTimer(self)
        self._search_timer.setSingleShot(True)
        self._search_timer.setInterval(300)  # delay of 300 ms
        self._search_timer.timeout.connect(self._execute_search)

        # start the timer each time the text changes
        self.textChanged.connect(self.on_text_changed)

    def clear_search_string(self):
        """Clear input"""
        self.clear()
        self._search_timer.stop()  # stop the timer to prevent triggering the search action
        self.window.controller.ctx.search_string_clear()

    def on_text_changed(self, text):
        """
        On text changed

        :param text: text entered by the user
        """
        self.clear_action.setVisible(bool(text))
        # restart the timer each time the text changes - ensures the action is only triggered
        # after a pause in typing
        self._search_timer.start()

    def _execute_search(self):
        """Invoke the search action after a specified delay."""
        search_text = self.text()
        self.window.controller.ctx.search_string_change(search_text)

    def focusInEvent(self, event):
        """
        Focus in event

        :param event: focus event
        """
        super(CtxSearchInput, self).focusInEvent(event)
        self.window.controller.ctx.search_focus_in()


#!/usr/bin/env python3
# -*- coding: utf-8 -*-
# ================================================== #
# This file is a part of PYGPT package               #
# Website: https://pygpt.net                         #
# GitHub:  https://github.com/szczyglis-dev/py-gpt   #
# MIT License                                        #
# Created By  : Marcin Szczygliński                  #
# Updated Date: 2025.07.21 21:00:00                  #
# ================================================== #

import json

from pygpt_net.core.types import (
    MODE_CHAT,
    MODE_VISION,
    MULTIMODAL_IMAGE,
    MODE_AUDIO,
    MULTIMODAL_AUDIO,
)

class ModelItem:

    OPENAI_COMPATIBLE = [
        "anthropic",
        "openai",
        "azure_openai",
        "google",
        "local_ai",
        "mistral_ai",
        "perplexity",
        "deepseek_api",
        "x_ai",
    ]

    def __init__(self, id=None):
        """
        Model data item

        :param id: Model ID
        """
        self.id = id
        self.name = None
        self.mode = ["chat"]
        self.langchain = {}
        self.llama_index = {}
        self.multimodal = ["text"]  # multimodal support: image, audio, etc.
        self.input = ["text"]  # multimodal support: image, audio, etc.
        self.output = ["text"]  # multimodal support: image, audio, etc.
        self.ctx = 0
        self.tokens = 0
        self.default = False
        self.imported = False
        self.provider = "openai"  # default provider
        self.tool_calls = False  # native tool calls available
        self.extra = {}

    def from_dict(self, data: dict):
        """
        Load data from dict

        :param data: dict
        """
        if 'id' in data:
            self.id = data['id']
        if 'name' in data:
            self.name = data['name']
        if 'mode' in data:
            mode = data['mode'].replace(' ', '')
            self.mode = mode.split(',')
        if 'input' in data:
            input = data['input'].replace(' ', '')
            self.input = input.split(',')
        if 'output' in data:
            output = data['output'].replace(' ', '')
            self.output = output.split(',')
        if 'ctx' in data:
            self.ctx = data['ctx']
        if 'tokens' in data:
            self.tokens = data['tokens']
        if 'default' in data:
            self.default = data['default']
        if 'extra' in data:
            self.extra = data['extra']
        if 'imported' in data:
            self.imported = data['imported']
        if 'provider' in data:
            self.provider = data['provider']
        if 'tool_calls' in data:
            self.tool_calls = data['tool_calls']

        # langchain
        """
        if 'langchain.provider' in data:
            self.langchain['provider'] = data['langchain.provider']
        if 'langchain.mode' in data:
            if data['langchain.mode'] is None or data['langchain.mode'] == "":
                self.langchain['mode'] = []
            else:
                mode = data['langchain.mode'].replace(' ', '')
                self.langchain['mode'] = mode.split(',')
        if 'langchain.args' in data:
            self.langchain['args'] = data['langchain.args']
        if 'langchain.env' in data:
            self.langchain['env'] = data['langchain.env']
        """
        
        # llama index
        if 'llama_index.provider' in data:
            self.llama_index['provider'] = data['llama_index.provider']  # backward compatibility < v2.5.20
        """
        if 'llama_index.mode' in data:
            if data['llama_index.mode'] is None or data['llama_index.mode'] == "":
                self.llama_index['mode'] = []
            else:
                mode = data['llama_index.mode'].replace(' ', '')
                self.llama_index['mode'] = mode.split(',')
        """
        if 'llama_index.args' in data:
            self.llama_index['args'] = data['llama_index.args']
        if 'llama_index.env' in data:
            self.llama_index['env'] = data['llama_index.env']

    def to_dict(self) -> dict:
        """
        Return data as dict

        :return: dict
        """
        data = {}
        data['id'] = self.id
        data['name'] = self.name
        data['mode'] = ','.join(self.mode)
        data['input'] = ','.join(self.input)
        data['output'] = ','.join(self.output)
        # data['langchain'] = self.langchain
        data['ctx'] = self.ctx
        data['tokens'] = self.tokens
        data['default'] = self.default
        data['extra'] = self.extra
        data['imported'] = self.imported
        data['provider'] = self.provider
        data['tool_calls'] = self.tool_calls

        # data['langchain.provider'] = None
        # data['langchain.mode'] = ""
        # data['langchain.args'] = []
        # data['langchain.env'] = []
        # data['llama_index.provider'] = None
        # data['llama_index.mode'] = ""
        data['llama_index.args'] = []
        data['llama_index.env'] = []


        # langchain
        """
        if 'provider' in self.langchain:
            data['langchain.provider'] = self.langchain['provider']        
        if 'mode' in self.langchain:
            data['langchain.mode'] = ",".join(self.langchain['mode'])
        if 'args' in self.langchain:
            # old versions support
            if isinstance(self.langchain['args'], dict):
                for key, value in self.langchain['args'].items():
                    item = {}
                    item['name'] = key
                    item['value'] = value
                    item['type'] = 'str'
                    data['langchain.args'].append(item)
            elif isinstance(self.langchain['args'], list):
                data['langchain.args'] = self.langchain['args']
        if 'env' in self.langchain:
            # old versions support
            if isinstance(self.langchain['env'], dict):
                for key, value in self.langchain['env'].items():
                    item = {}
                    item['name'] = key
                    item['value'] = value
                    data['langchain.env'].append(item)
            elif isinstance(self.langchain['env'], list):
                data['langchain.env'] = self.langchain['env']
        """

        # llama_index
        # if 'provider' in self.llama_index:
            # data['llama_index.provider'] = self.llama_index['provider']
        # if 'mode' in self.llama_index:
            # data['llama_index.mode'] = ",".join(self.llama_index['mode'])
        if 'args' in self.llama_index:
            # old versions support
            if isinstance(self.llama_index['args'], dict):
                for key, value in self.llama_index['args'].items():
                    item = {}
                    item['name'] = key
                    item['value'] = value
                    item['type'] = 'str'
                    data['llama_index.args'].append(item)
            elif isinstance(self.llama_index['args'], list):
                data['llama_index.args'] = self.llama_index['args']
        if 'env' in self.llama_index:
            # old versions support
            if isinstance(self.llama_index['env'], dict):
                for key, value in self.llama_index['env'].items():
                    item = {}
                    item['name'] = key
                    item['value'] = value
                    data['llama_index.env'].append(item)
            elif isinstance(self.llama_index['env'], list):
                data['llama_index.env'] = self.llama_index['env']

        return data

    def is_supported(self, mode: str) -> bool:
        """
        Check if model supports mode

        :param mode: Mode
        :return: True if supported
        """
        if mode == MODE_CHAT and not self.is_openai_supported():
            # only OpenAI API compatible models are supported in Chat mode
            return False
        return mode in self.mode

    def is_multimodal(self) -> bool:
        """
        Check if model is multimodal

        :return: True if multimodal
        """
        return len(self.multimodal) > 0

    def is_openai_supported(self) -> bool:
        """
        Check if model is supported by OpenAI API (or compatible)

        :return: True if OpenAI compatible
        """
        return self.provider in self.OPENAI_COMPATIBLE

    def is_gpt(self) -> bool:
        """
        Check if model is supported by OpenAI Responses API

        :return: True if OpenAI Responses API compatible
        """
        if (self.id.startswith("gpt-")
                or self.id.startswith("chatgpt")
                or self.id.startswith("o1")
                or self.id.startswith("o3")
                or self.id.startswith("o4")
                or self.id.startswith("o5")
                or self.id.startswith("codex-")
                or self.id.startswith("dall-e-")):
            return True
        return False

    def is_ollama(self) -> bool:
        """
        Check if model is Ollama

        :return: True if Ollama
        """
        if self.provider == "ollama":
            return True
        if self.llama_index is None:
            return False
        if self.llama_index.get("provider") is None:
            return False
        return "ollama" in self.llama_index.get("provider", "")

    def get_provider(self):
        return self.provider

    def get_ollama_model(self) -> str:
        """
        Get Ollama model ID

        :return: model ID
        """
        if "args" in self.llama_index:
            for arg in self.llama_index["args"]:
                if arg["name"] == "model":
                    return arg["value"]
        return ""

    def has_mode(self, mode: str) -> bool:
        """
        Check if model has mode

        :param mode: Mode
        :return: True if supported
        """
        return mode in self.mode

    def add_mode(self, mode: str):
        """
        Add mode

        :param mode: Mode
        """
        if mode not in self.mode:
            self.mode.append(mode)

    def remove_mode(self, mode: str):
        """
        Remove mode

        :param mode: Mode
        """
        if mode in self.mode:
            self.mode.remove(mode)

    def is_image_input(self) -> bool:
        """
        Check if model supports image input

        :return: True if supports image input
        """
        if MODE_VISION in self.mode or MULTIMODAL_IMAGE in self.input:
            return True
        return False

    def is_audio_input(self) -> bool:
        """
        Check if model supports audio input

        :return: True if supports audio input
        """
        if MODE_AUDIO in self.mode or MULTIMODAL_AUDIO in self.input:
            return True
        return False

    def dump(self) -> str:
        """
        Dump event to json string

        :return: JSON string
        """
        try:
            return json.dumps(self.to_dict())
        except Exception as e:
            pass
        return ""

    def __str__(self) -> str:
        """
        To string

        :return: Dumped JSON string
        """
        return self.dump()



#!/usr/bin/env python3
# -*- coding: utf-8 -*-
# ================================================== #
# This file is a part of PYGPT package               #
# Website: https://pygpt.net                         #
# GitHub:  https://github.com/szczyglis-dev/py-gpt   #
# MIT License                                        #
# Created By  : Marcin Szczygliński                  #
# Updated Date: 2025.07.20 23:00:00                  #
# ================================================== #

import datetime

from PySide6 import QtWidgets, QtCore, QtGui
from PySide6.QtCore import Qt, QPoint, QItemSelectionModel
from PySide6.QtGui import QAction, QIcon, QColor, QPixmap, QStandardItem
from PySide6.QtWidgets import QMenu
from overrides import overrides

from pygpt_net.ui.widget.lists.base import BaseList
from pygpt_net.utils import trans
import pygpt_net.icons_rc


class ContextList(BaseList):
    def __init__(self, window=None, id=None):
        """
        Context select menu

        :param window: main window
        :param id: input id
        """
        super(ContextList, self).__init__(window)
        self.window = window
        self.id = id
        self.expanded_items = set()
        self.setItemDelegate(ImportantItemDelegate())
        self.setContextMenuPolicy(Qt.CustomContextMenu)
        self.customContextMenuRequested.connect(self.show_context_menu)
        self._backup_selection = None
        self.restore_after_ctx_menu = True

    def click(self, index):
        """
        Click event (override, connected in BaseList class)

        :param index: index
        """
        # only if left mouse button:
        item = self.window.ui.models['ctx.list'].itemFromIndex(index)
        if item is not None:
            if not hasattr(item, 'isFolder'):
                return
            if item.isFolder:
                self.window.controller.ctx.set_group(item.id)
                if self.window.ui.nodes['ctx.list'].isExpanded(index):
                    self.expanded_items.discard(item.id)
                    self.window.ui.nodes['ctx.list'].collapse(index)
                else:
                    self.expanded_items.add(item.id)
                    self.window.ui.nodes['ctx.list'].expand(index)
            else:
                self.window.controller.ctx.select_by_id(item.id)
        else:
            pass  # unselected

    def expand_group(self, id):
        """
        Expand group

        :param id: group id
        """
        for i in range(self.window.ui.models['ctx.list'].rowCount()):
            item = self.window.ui.models['ctx.list'].item(i)
            if isinstance(item, GroupItem):
                if item.id == id:
                    index = self.window.ui.models['ctx.list'].indexFromItem(item)
                    self.window.ui.nodes['ctx.list'].expand(index)
                    self.expanded_items.add(id)

    def dblclick(self, index):
        """
        Double click event

        :param index: index
        """
        print("dblclick")

    def mousePressEvent(self, event):
        if event.button() == Qt.LeftButton:
            index = self.indexAt(event.pos())
            if not index.isValid():
                self.window.controller.ctx.unselect()
                return
            super().mousePressEvent(event)
        elif event.button() == Qt.RightButton:
            index = self.indexAt(event.pos())
            if index.isValid():
                self._backup_selection = list(self.selectionModel().selectedIndexes())
                self.selectionModel().clearSelection()
                self.selectionModel().select(
                    index, QItemSelectionModel.Select | QItemSelectionModel.Rows
                )
            event.accept()
        else:
            super().mousePressEvent(event)

    def show_context_menu(self, pos: QPoint):
        """
        Context menu event

        :param pos: QPoint
        """
        global_pos = self.viewport().mapToGlobal(pos)
        index = self.indexAt(pos)
        item = self.window.ui.models['ctx.list'].itemFromIndex(index)

        if (item is not None
                and index.isValid()
                and hasattr(item, 'id')):

            idx = item.row()
            id = item.id

            # group context menu
            if hasattr(item, 'isFolder') and item.isFolder:
                actions = {}
                actions['new'] = QAction(QIcon(":/icons/add.svg"), trans('action.ctx.new'), self)
                actions['new'].triggered.connect(
                    lambda checked=False, id=id: self.window.controller.ctx.new(force=False, group_id=id)
                )
                actions['rename'] = QAction(QIcon(":/icons/edit.svg"), trans('action.rename'), self)
                actions['rename'].triggered.connect(
                    lambda checked=False, id=id: self.window.controller.ctx.rename_group(id)
                )
                actions['delete'] = QAction(QIcon(":/icons/delete.svg"), trans('action.group.delete.only'), self)
                actions['delete'].triggered.connect(
                    lambda checked=False, id=id: self.window.controller.ctx.delete_group(id)
                )
                actions['delete_all'] = QAction(QIcon(":/icons/delete.svg"), trans('action.group.delete.all'), self)
                actions['delete_all'].triggered.connect(
                    lambda checked=False, id=id: self.window.controller.ctx.delete_group_all(id)
                )

                menu = QMenu(self)
                menu.addAction(actions['new'])
                menu.addAction(actions['rename'])
                menu.addAction(actions['delete'])  # delete group
                menu.addAction(actions['delete_all'])  # delete group and all contexts

                if idx >= 0:
                    menu.exec_(global_pos)

            # children context menu
            else:
                ctx_id = id
                ctx = self.window.core.ctx.get_meta_by_id(id)
                if ctx is None:
                    return

                is_important = ctx.important

                actions = {}
                actions['open'] = QAction(QIcon(":/icons/chat.svg"), trans('action.open'), self)
                actions['open'].triggered.connect(
                    lambda checked=False, ctx_id=ctx_id: self.action_open(ctx_id, idx)
                )

                actions['open_new_tab'] = QAction(QIcon(":/icons/chat.svg"), trans('action.open_new_tab'), self)
                actions['open_new_tab'].triggered.connect(
                    lambda checked=False, ctx_id=ctx_id: self.action_open_new_tab(ctx_id, idx)
                )

                actions['rename'] = QAction(QIcon(":/icons/edit.svg"), trans('action.rename'), self)
                actions['rename'].triggered.connect(
                    lambda checked=False, ctx_id=ctx_id: self.action_rename(ctx_id)
                )

                if is_important:
                    actions['important'] = QAction(QIcon(":/icons/pin.svg"), trans('action.unpin'), self)
                    actions['important'].triggered.connect(
                        lambda checked=False, ctx_id=ctx_id: self.action_unpin(ctx_id)
                    )
                else:
                    actions['important'] = QAction(QIcon(":/icons/pin.svg"), trans('action.pin'), self)
                    actions['important'].triggered.connect(
                        lambda checked=False, ctx_id=ctx_id: self.action_pin(ctx_id)
                    )

                actions['duplicate'] = QAction(QIcon(":/icons/copy.svg"), trans('action.duplicate'), self)
                actions['duplicate'].triggered.connect(
                    lambda checked=False, ctx_id=ctx_id: self.action_duplicate(ctx_id)
                )

                actions['delete'] = QAction(QIcon(":/icons/delete.svg"), trans('action.delete'), self)
                actions['delete'].triggered.connect(
                    lambda checked=False, ctx_id=ctx_id: self.action_delete(ctx_id)
                )

                actions['copy_id'] = QAction(QIcon(":/icons/copy.svg"), trans('action.ctx_copy_id') + " @" + str(id), self)
                actions['copy_id'].triggered.connect(
                    lambda checked=False, ctx_id=ctx_id: self.action_copy_id(ctx_id)
                )

                actions['reset'] = QAction(QIcon(":/icons/close.svg"), trans('action.ctx_reset'), self)
                actions['reset'].triggered.connect(
                    lambda checked=False, ctx_id=ctx_id: self.action_reset(ctx_id)
                )

                menu = QMenu(self)
                menu.addAction(actions['open'])
                menu.addAction(actions['open_new_tab'])
                menu.addAction(actions['rename'])
                menu.addAction(actions['duplicate'])
                menu.addAction(actions['important'])
                menu.addAction(actions['delete'])

                # set label menu
                colors = self.window.controller.ui.get_colors()
                set_label_menu = menu.addMenu(trans('calendar.day.label'))
                for status_id, status_info in colors.items():
                    name = trans('calendar.day.' + status_info['label'])
                    if status_id == 0:
                        name = '-'
                    color = status_info['color']
                    pixmap = QPixmap(16, 16)
                    pixmap.fill(color)
                    icon = QIcon(pixmap)
                    status_action = QAction(icon, name, self)
                    status_action.triggered.connect(
                        lambda checked=False,
                               ctx_id=ctx_id,
                               s_id=status_id: self.window.controller.ctx.set_label(ctx_id, s_id)
                    )
                    set_label_menu.addAction(status_action)

                idx_menu = QMenu(trans('action.idx'), self)

                # indexes list
                idxs = self.window.core.config.get('llama.idx.list')
                store = self.window.core.idx.get_current_store()  # get current idx store provider
                if len(idxs) > 0:
                    for index in idxs:
                        id = index['id']
                        name = index['name'] + " (" + index['id'] + ")"

                        # add to index
                        action = idx_menu.addAction("IDX: " + name)
                        action.setIcon(QIcon(":/icons/db.svg"))
                        action.triggered.connect(
                            lambda checked=False,
                                   ctx_id=ctx_id,
                                   index=id: self.action_idx(ctx_id, index)
                        )

                    # remove from index
                    if ctx.indexed is not None and ctx.indexed > 0:
                        # get list of indexes in which context is indexed
                        if store in ctx.indexes:
                            store_indexes = ctx.indexes[store]
                            idx_menu.addSeparator()
                            for store_index in store_indexes:
                                action = idx_menu.addAction(trans("action.idx.remove") + ": " + store_index)
                                action.setIcon(QIcon(":/icons/delete.svg"))
                                action.triggered.connect(
                                    lambda checked=False,
                                           store_index=store_index,
                                           ctx_id=ctx_id: self.action_idx_remove(store_index, ctx_id)  # by context meta id
                                )
                    menu.addMenu(idx_menu)

                # -----------------------------------------

                # move to group menu
                group_menu = QMenu(trans('action.move_to'), self)
                groups = self.window.core.ctx.get_groups()

                # add group
                action = group_menu.addAction(trans("action.group.new"))
                action.setIcon(QIcon(":/icons/add.svg"))
                action.triggered.connect(
                    lambda checked=False, ctx_id=ctx_id: self.window.controller.ctx.new_group(ctx_id)
                )

                # add separator if groups exists
                if len(groups) > 0:
                    group_menu.addSeparator()

                # list of groups
                for group_id in groups:
                    group = groups[group_id]
                    action = group_menu.addAction(group.name)
                    action.setIcon(QIcon(":/icons/folder_filled.svg"))
                    action.triggered.connect(
                        lambda checked=False,
                               group_id=group_id,
                               ctx_id=ctx_id: self.window.controller.ctx.move_to_group(ctx_id, group_id)
                    )

                # add separator if groups exists
                if len(groups) > 0:
                    group_menu.addSeparator()

                # if in group add remove from group
                if ctx.group_id is not None and ctx.group_id > 0:
                    group_name = str(ctx.group_id)
                    if ctx.group_id in groups:
                        group_name = groups[ctx.group_id].name
                    action = group_menu.addAction(trans("action.group.remove") + ": " + group_name)
                    action.setIcon(QIcon(":/icons/delete.svg"))
                    action.triggered.connect(
                        lambda checked=False,
                               ctx_id=ctx_id: self.window.controller.ctx.remove_from_group(ctx_id)
                    )

                menu.addMenu(group_menu)

                menu.addAction(actions['copy_id'])

                # -----------------------------------------

                # show last indexed date if available
                if ctx.indexed is not None and ctx.indexed > 0:
                    suffix = ""
                    if ctx.updated > ctx.indexed:
                        suffix = " *"
                    dt = datetime.datetime.fromtimestamp(ctx.indexed).strftime("%Y-%m-%d %H:%M")
                    action = QAction(QIcon(":/icons/clock.svg"), trans('action.ctx.indexed') + ": " + dt + suffix, self)
                    action.setEnabled(False)  # disable action, only for info
                    menu.addAction(action)

                menu.addAction(actions['reset'])

                if idx >= 0:
                    self.window.controller.ctx.set_selected(ctx_id)
                    menu.exec_(global_pos)

        # store previous scroll position
        self.store_scroll_position()

        # restore selection if it was backed up
        if self.restore_after_ctx_menu:
            if self._backup_selection is not None:
                self.selectionModel().clearSelection()
                for idx in self._backup_selection:
                    self.selectionModel().select(
                        idx, QItemSelectionModel.Select | QItemSelectionModel.Rows
                    )
                self._backup_selection = None

        # restore scroll position
        self.restore_after_ctx_menu = True
        self.restore_scroll_position()

    def action_open(self, id: int, idx: int = None):
        """
        Open context action handler

        :param id: context id
        :param idx: index id (optional)
        """
        self.restore_after_ctx_menu = False  # do not restore selection after context menu
        self.window.controller.ctx.load(id, select_idx=idx)

    def action_open_new_tab(self, id: int, idx: int = None):
        """
        Open context action handler

        :param id: context id
         :param idx: index id (optional)
        """
        self.restore_after_ctx_menu = False  # do not restore selection after context menu
        self.window.controller.ctx.load(id, select_idx=idx, new_tab=True)

    def action_idx(self, id: int, idx: int):
        """
        Index with llama context action handler

        :param id: context id
        :param idx: index name
        """
        self.restore_after_ctx_menu = False  # do not restore selection after context menu
        self.window.controller.idx.indexer.index_ctx_meta(id, idx)

    def action_idx_remove(self, idx: str, meta_id: int):
        """
        Remove from index action handler

        :param idx: index id
        :param meta_id: meta id
        """
        self.restore_after_ctx_menu = False  # do not restore selection after context menu
        self.window.controller.idx.indexer.index_ctx_meta_remove(idx, meta_id)

    def action_rename(self, id):
        """
        Rename action handler

        :param id: context id
        """
        self.restore_after_ctx_menu = False  # do not restore selection after context menu
        self.window.controller.ctx.rename(id)

    def action_pin(self, id):
        """
        Pin action handler

        :param id: context id
        """
        self.restore_after_ctx_menu = False  # do not restore selection after context menu
        self.window.controller.ctx.set_important(id, True)

    def action_unpin(self, id):
        """
        Unpin action handler

        :param id: context id
        """
        self.restore_after_ctx_menu = False  # do not restore selection after context menu
        self.window.controller.ctx.set_important(id, False)

    def action_important(self, id):
        """
        Set as important action handler

        :param id: context id
        """
        self.restore_after_ctx_menu = False  # do not restore selection after context menu
        self.window.controller.ctx.set_important(id)

    def action_duplicate(self, id):
        """
        Rename duplicate handler

        :param id: context id
        """
        self.window.controller.ctx.common.duplicate(id)

    def action_delete(self, id):
        """
        Delete action handler

        :param id: context id
        """
        self.restore_after_ctx_menu = False  # do not restore selection after context menu
        self.window.controller.ctx.delete(id)

    def action_copy_id(self, id):
        """
        Copy ID tag action handler

        :param id: context id
        """
        self.window.controller.ctx.common.copy_id(id)

    def action_reset(self, id):
        """
        Reset action handler

        :param id: context id
        """
        self.restore_after_ctx_menu = False  # do not restore selection after context menu
        self.window.controller.ctx.common.reset(id)


    def selectionCommand(self, index, event=None):
        """
        Selection command
        :param index: Index
        :param event: Event
        """
        return super().selectionCommand(index, event)


class ImportantItemDelegate(QtWidgets.QStyledItemDelegate):
    """
    Label color delegate

    :param QtWidgets.QStyledItemDelegate: parent class
    """
    def paint(self, painter, option, index):
        if index.parent().isValid():
            option.rect.adjust(15, 0, 0, 0)

        super(ImportantItemDelegate, self).paint(painter, option, index)

        # pin (>= 10)
        if index.data(QtCore.Qt.ItemDataRole.UserRole):
            data = index.data(QtCore.Qt.ItemDataRole.UserRole)
            label = 0
            is_important = False
            is_attachment = False
            is_group = False
            in_group = False

            if "label" in data:
                label = data["label"]
            if "is_important" in data and data["is_important"]:
                is_important = True
            if "is_attachment" in data and data["is_attachment"]:
                is_attachment = True
            if "is_group" in data and data["is_group"]:
                is_group = True
            if "in_group" in data and data["in_group"]:
                in_group = True

            painter.save()

            if is_attachment:
                icon = QtGui.QIcon(":/icons/attachment.svg")
                icon_size = option.decorationSize or QtCore.QSize(16, 16)
                icon_pos = option.rect.right() - icon_size.width()
                icon_rect = QtCore.QRect(
                    icon_pos,
                    option.rect.top() + (option.rect.height() - icon_size.height()) / 2,
                    icon_size.width(),
                    icon_size.height()
                )
                icon.paint(painter, icon_rect, QtCore.Qt.AlignCenter)

            if is_important:
                color = self.get_color_for_status(3)
                square_size = 3
                square_margin = 0
                square_rect = QtCore.QRect(
                    option.rect.left() + square_margin,
                    option.rect.top() + 2,
                    square_size,
                    square_size,
                )
                painter.setBrush(color)
                painter.setPen(
                    QtGui.QPen(
                        QtCore.Qt.black,
                        0.5,
                        QtCore.Qt.SolidLine,
                    )
                )
                painter.drawRect(square_rect)

            # label (0-9)
            if label > 0:
                color = self.get_color_for_status(label)
                square_size = 5
                square_margin = 0
                square_rect = QtCore.QRect(
                    option.rect.left() + square_margin,
                    option.rect.center().y() - (square_size / 2) + 2,
                    square_size,
                    square_size,
                )
                painter.setBrush(color)
                painter.setPen(QtCore.Qt.NoPen)
                painter.drawRect(square_rect)

            painter.restore()

    def get_color_for_status(self, status: int) -> QColor:
        """
        Get color for status

        :param status: status id
        :return: color
        """
        statuses = {
            0: {'label': 'label.color.default', 'color': QColor(100, 100, 100)},
            1: {'label': 'label.color.red', 'color': QColor(255, 0, 0)},
            2: {'label': 'label.color.orange', 'color': QColor(255, 165, 0)},
            3: {'label': 'label.color.yellow', 'color': QColor(255, 255, 0)},
            4: {'label': 'label.color.green', 'color': QColor(0, 255, 0)},
            5: {'label': 'label.color.blue', 'color': QColor(0, 0, 255)},
            6: {'label': 'label.color.indigo', 'color': QColor(75, 0, 130)},
            7: {'label': 'label.color.violet', 'color': QColor(238, 130, 238)},
        }
        if status in statuses:
            return statuses[status]['color']
        else:
            return statuses[0]['color']

class GroupItem(QStandardItem):
    def __init__(self, icon, name, id):
        super().__init__(icon, name)
        self.id = id
        self.name = name
        self.isFolder = True
        self.isPinned = False
        self.hasAttachments = False
        self.dt = None

class Item(QStandardItem):
    def __init__(self, name, id):
        super().__init__(name)
        self.id = id
        self.name = name
        self.isFolder = False
        self.isPinned = False
        self.dt = None

class SectionItem(QStandardItem):
    def __init__(self, title, group: bool = False):
        super().__init__(title)
        self.title = title
        self.group = group
        self.setSelectable(False)
        self.setEnabled(False)
        if self.group:
            self.setTextAlignment(QtCore.Qt.AlignRight)
        else:
            self.setTextAlignment(QtCore.Qt.AlignRight)
        font = self.font()
        font.setBold(True)
        self.setFont(font)



#!/usr/bin/env python3
# -*- coding: utf-8 -*-
# ================================================== #
# This file is a part of PYGPT package               #
# Website: https://pygpt.net                         #
# GitHub:  https://github.com/szczyglis-dev/py-gpt   #
# MIT License                                        #
# Created By  : Marcin Szczygliński                  #
# Updated Date: 2025.07.19 00:00:00                  #
# ================================================== #

import base64
from typing import Optional

from PySide6.QtCore import QObject, Signal, Slot, QRunnable

from pygpt_net.core.bridge import BridgeContext
from pygpt_net.core.events import RenderEvent
from pygpt_net.core.types import MODE_ASSISTANT
from pygpt_net.item.ctx import CtxItem


class StreamWorker(QObject, QRunnable):
    end = Signal(object)
    errorOccurred = Signal(Exception)
    eventReady = Signal(object)

    def __init__(self, ctx: CtxItem, window, parent=None):
        QObject.__init__(self)
        QRunnable.__init__(self)
        self.ctx = ctx
        self.window = window

    @Slot()
    def run(self):
        output = ""
        output_tokens = 0
        begin = True
        error = None
        fn_args_buffers = {}
        citations = []
        files = []
        img_path = self.window.core.image.gen_unique_path(self.ctx)
        is_image = False
        is_code = False

        data = {
            "meta": self.ctx.meta,
            "ctx": self.ctx
        }
        event = RenderEvent(RenderEvent.STREAM_BEGIN, data)
        self.eventReady.emit(event)

        try:
            if self.ctx.stream is not None:
                tool_calls = []
                for chunk in self.ctx.stream:
                    # if force stop then break
                    if self.window.controller.kernel.stopped():
                        break
                    if error is not None:
                        break  # break if error

                    etype = None
                    response = None
                    chunk_type = "raw"

                    if self.ctx.use_responses_api:
                        if hasattr(chunk, 'type'):  # streaming event type
                            etype = chunk.type
                            chunk_type = "api_chat_responses"  # responses API
                        else:
                            continue
                    else:
                        if (hasattr(chunk, 'choices')
                                and chunk.choices[0] is not None
                                and hasattr(chunk.choices[0], 'delta')
                                and chunk.choices[0].delta is not None):
                            chunk_type = "api_chat"  # chat completions API
                        elif (hasattr(chunk, 'choices')
                              and chunk.choices[0] is not None
                              and hasattr(chunk.choices[0], 'text')
                              and chunk.choices[0].text is not None):
                            chunk_type = "api_completion"
                        elif (hasattr(chunk, 'content')
                              and chunk.content is not None):
                            chunk_type = "langchain_chat"
                        elif (hasattr(chunk, 'delta')
                              and chunk.delta is not None):
                            chunk_type = "llama_chat"

                    # OpenAI chat completion
                    if chunk_type == "api_chat":
                        citations = None
                        if chunk.choices[0].delta and chunk.choices[0].delta.content is not None:
                            if citations is None:
                                if chunk and hasattr(chunk, 'citations') and chunk.citations is not None:
                                    citations = chunk.citations
                                    self.ctx.urls = citations
                            response = chunk.choices[0].delta.content

                        if chunk.choices[0].delta and chunk.choices[0].delta.tool_calls:
                            tool_chunks = chunk.choices[0].delta.tool_calls
                            for tool_chunk in tool_chunks:
                                if tool_chunk.index is None:
                                    tool_chunk.index = 0
                                if len(tool_calls) <= tool_chunk.index:
                                    tool_calls.append(
                                        {
                                            "id": "",
                                            "type": "function",
                                            "function": {
                                                "name": "",
                                                "arguments": ""
                                            }
                                        }
                                    )
                                tool_call = tool_calls[tool_chunk.index]
                                if tool_chunk.id:
                                    tool_call["id"] += tool_chunk.id
                                if tool_chunk.function.name:
                                    tool_call["function"]["name"] += tool_chunk.function.name
                                if tool_chunk.function.arguments:
                                    tool_call["function"]["arguments"] += tool_chunk.function.arguments

                    # OpenAI Responses API
                    elif chunk_type == "api_chat_responses":
                        if etype == "response.output_text.delta":
                            response = chunk.delta

                        # ---------- function_call ----------
                        elif etype == "response.output_item.added" and chunk.item.type == "function_call":
                            tool_calls.append({
                                "id": chunk.item.id,
                                "call_id": chunk.item.call_id,
                                "type": "function",
                                "function": {
                                    "name": chunk.item.name,
                                    "arguments": ""
                                }
                            })
                            fn_args_buffers[chunk.item.id] = ""
                        elif etype == "response.function_call_arguments.delta":
                            fn_args_buffers[chunk.item_id] += chunk.delta
                        elif etype == "response.function_call_arguments.done":
                            for tc in tool_calls:
                                if tc["id"] == chunk.item_id:
                                    tc["function"]["arguments"] = fn_args_buffers[chunk.item_id]
                                    break
                            fn_args_buffers.pop(chunk.item_id, None)

                        # ---------- annotations ----------
                        elif etype == "response.output_text.annotation.added":
                            if chunk.annotation['type'] == "url_citation":
                                if citations is None:
                                    citations = []
                                url_citation = chunk.annotation['url']
                                citations.append(url_citation)
                                self.ctx.urls = citations
                            if chunk.annotation['type'] == "container_file_citation":
                                container_id = chunk.annotation['container_id']
                                file_id = chunk.annotation['file_id']
                                files.append({
                                    "container_id": container_id,
                                    "file_id": file_id,
                                })

                        # ---------- code interpreter ----------
                        elif etype == "response.code_interpreter_call_code.delta":
                            if not is_code:
                                response = "\n\n**Code interpreter**\n```python\n" + chunk.delta
                                is_code = True
                            else:
                                response = chunk.delta
                        elif etype == "response.code_interpreter_call_code.done":
                            response = "\n\n```\n-----------\n"

                        # ---------- image gen ----------
                        elif etype == "response.image_generation_call.partial_image":
                            image_base64 = chunk.partial_image_b64
                            image_bytes = base64.b64decode(image_base64)
                            with open(img_path, "wb") as f:
                                f.write(image_bytes)
                            is_image = True

                        # ---------- response ID ----------
                        elif etype == "response.created":
                            self.ctx.msg_id = chunk.response.id

                        # ---------- end / error ----------
                        elif etype in {"response.done", "response.failed", "error"}:
                            pass

                    # OpenAI completion
                    elif chunk_type == "api_completion":
                        if chunk.choices[0].text is not None:
                            response = chunk.choices[0].text

                    # langchain chat
                    elif chunk_type == "langchain_chat":
                        if chunk.content is not None:
                            response = str(chunk.content)

                    # llama chat
                    elif chunk_type == "llama_chat":
                        if chunk.delta is not None:
                            response = str(chunk.delta)
                        tool_chunks = chunk.message.additional_kwargs.get("tool_calls", [])
                        if tool_chunks:
                            for tool_chunk in tool_chunks:
                                id_val = None
                                name = None
                                args = {}
                                if hasattr(tool_chunk, 'arguments'):
                                    args = tool_chunk.arguments
                                elif hasattr(tool_chunk, 'function') and hasattr(tool_chunk.function, 'arguments'):
                                    args = tool_chunk.function.arguments
                                if hasattr(tool_chunk, 'call_id'):
                                    id_val = tool_chunk.call_id
                                elif hasattr(tool_chunk, 'id'):
                                    id_val = tool_chunk.id
                                if hasattr(tool_chunk, 'name'):
                                    name = tool_chunk.name
                                elif hasattr(tool_chunk, 'function') and hasattr(tool_chunk.function, 'name'):
                                    name = tool_chunk.function.name
                                if id_val:
                                    if not args:
                                        args = "{}"  # JSON encoded
                                    tool_call = {
                                        "id": id_val,
                                        "type": "function",
                                        "function": {
                                            "name": name,
                                            "arguments": args
                                        }
                                    }
                                    tool_calls.clear()
                                    tool_calls.append(tool_call)

                    # raw text: llama-index and langchain completion
                    else:
                        if chunk is not None:
                            response = str(chunk)

                    if response is not None and response != "":
                        if begin and response == "":  # prevent empty beginning
                            continue
                        output += response
                        output_tokens += 1
                        data = {
                            "meta": self.ctx.meta,
                            "ctx": self.ctx,
                            "chunk": response,
                            "begin": begin,
                        }
                        event = RenderEvent(RenderEvent.STREAM_APPEND, data)
                        self.eventReady.emit(event)
                        begin = False

                # unpack and store tool calls
                if tool_calls:
                    self.window.core.debug.info("[chat] Tool calls found, unpacking...")
                    self.window.core.command.unpack_tool_calls_chunks(self.ctx, tool_calls)

                # append images
                if is_image:
                    self.window.core.debug.info("[chat] Image generation call found")
                    self.ctx.images = [img_path]  # save image path to ctx

        except Exception as e:
            error = e

        # update ctx
        self.ctx.output = output
        self.ctx.set_tokens(self.ctx.input_tokens, output_tokens)

        # if files from container are found, download them and append to ctx
        if files:
            self.window.core.debug.info("[chat] Container files found, downloading...")
            try:
                self.window.core.gpt.container.download_files(self.ctx, files)
            except Exception as e:
                self.window.core.debug.error(f"[chat] Error downloading container files: {e}")

        self.end.emit(self.ctx)
        if error:
            self.errorOccurred.emit(error)


class Stream:
    def __init__(self, window=None):
        """
        Stream controller

        :param window: Window instance
        """
        self.window = window
        self.ctx = None
        self.mode = None
        self.thread = None
        self.worker = None
        self.is_response = False
        self.reply = False
        self.internal = False
        self.context = None
        self.extra = {}

    def append(
            self,
            ctx: CtxItem,
            mode: str = None,
            is_response: bool = False,
            reply: str = False,
            internal: bool = False,
            context: Optional[BridgeContext] = None,
            extra: Optional[dict] = None
    ):
        """
        Asynchronous append of stream worker to the thread.

        :param ctx: CtxItem
        :param mode: Mode of stream processing
        :param is_response: Is this a response stream?
        :param reply: Reply text
        :param internal: Internal flag for handling
        :param context: Optional BridgeContext for additional context
        :param extra: Optional extra data for the stream
        """
        self.ctx = ctx
        self.mode = mode
        self.is_response = is_response
        self.reply = reply
        self.internal = internal
        self.context = context
        self.extra = extra if extra is not None else {}

        self.worker = StreamWorker(ctx, self.window)
        self.worker.eventReady.connect(self.handleEvent)
        self.worker.errorOccurred.connect(self.handleError)
        self.worker.end.connect(self.handleEnd)

        self.window.core.debug.info("[chat] Stream begin...")
        self.window.threadpool.start(self.worker)

    @Slot(object)
    def handleEnd(self, ctx: CtxItem):
        """
        Slot for handling end of stream

        This method is called when the stream processing is finished.

        :param ctx: CtxItem
        """
        self.window.controller.ui.update_tokens()

        data = {"meta": self.ctx.meta, "ctx": self.ctx}
        event = RenderEvent(RenderEvent.STREAM_END, data)
        self.window.dispatch(event)
        self.window.controller.chat.output.handle_after(
            ctx=ctx,
            mode=self.mode,
            stream=True,
        )

        # finish: assistant thread run
        if self.mode == MODE_ASSISTANT:
            self.window.controller.assistant.threads.handle_output_message_after_stream(ctx)
        else:
            # finish: KernelEvent.RESPONSE_OK, KernelEvent.RESPONSE_ERROR
            if self.is_response:
                # post-handle, execute cmd, etc.
                self.window.controller.chat.response.post_handle(
                    ctx=ctx,
                    mode=self.mode,
                    stream=True,
                    reply=self.reply,
                    internal=self.internal
                )

    def handleEvent(self, event):
        """
        Slot for handling RenderEvent

        :param event: RenderEvent
        """
        self.window.dispatch(event)

    def handleError(self, error):
        """
        Slot for handling errors

        :param error: Exception
        """
        self.window.core.debug.log(error)
        if self.is_response:
            if not isinstance(self.extra, dict):
                self.extra = {}
            self.extra["error"] = error
            self.window.controller.chat.response.failed(self.context, self.extra) # send error
            # post-handle, execute cmd, etc.
            self.window.controller.chat.response.post_handle(
                ctx=self.ctx,
                mode=self.mode,
                stream=True,
                reply=self.reply,
                internal=self.internal,
            )

    def log(self, data: object):
        """
        Save debug log data

        :param data: log data
        """
        self.window.core.debug.info(data)


#!/usr/bin/env python3
# -*- coding: utf-8 -*-
# ================================================== #
# This file is a part of PYGPT package               #
# Website: https://pygpt.net                         #
# GitHub:  https://github.com/szczyglis-dev/py-gpt   #
# MIT License                                        #
# Created By  : Marcin Szczygliński                  #
# Updated Date: 2025.07.22 22:00:00                  #
# ================================================== #

from typing import Any
from pygpt_net.item.model import ModelItem
from pygpt_net.item.ctx import CtxItem

class Response:
    def __init__(self, window=None):
        """
        Response

        :param window: Window instance
        """
        self.window = window

    def from_react(
            self,
            ctx: CtxItem,
            model: ModelItem,
            response: Any
    ) -> None:
        pass

    def from_index(
            self,
            ctx: CtxItem,
            model: ModelItem,
            response: Any
    ) -> None:
        output = str(response.response)
        if output is None:
            output = ""
        ctx.set_output(output, "")

    def from_llm(
            self,
            ctx: CtxItem,
            model: ModelItem,
            llm,
            response: Any
    ) -> None:
        output = response.message.content
        tool_calls = llm.get_tool_calls_from_response(
            response,
            error_on_no_tool_call=False,
        )
        if output is None:
            output = ""
        ctx.set_output(output, "")
        ctx.tool_calls = self.window.core.command.unpack_tool_calls_from_llama(tool_calls)

    def from_index_stream(
            self,
            ctx: CtxItem,
            model: ModelItem,
            response: Any
    ) -> None:
        ctx.stream = response.response_gen
        ctx.set_output("", "")

    def from_llm_stream(
            self,
            ctx: CtxItem,
            model: ModelItem,
            llm,
            response: Any
    ) -> None:
        ctx.stream = response  # chunk is in response.delta
        ctx.set_output("", "")


#!/usr/bin/env python3
# -*- coding: utf-8 -*-
# ================================================== #
# This file is a part of PYGPT package               #
# Website: https://pygpt.net                         #
# GitHub:  https://github.com/szczyglis-dev/py-gpt   #
# MIT License                                        #
# Created By  : Marcin Szczygliński                  #
# Updated Date: 2025.07.19 17:00:00                  #
# ================================================== #

from PySide6.QtCore import QItemSelectionModel, QPoint
from PySide6.QtGui import QAction, QIcon, Qt
from PySide6.QtWidgets import QMenu

from pygpt_net.ui.widget.lists.base import BaseList
from pygpt_net.utils import trans
import pygpt_net.icons_rc


class AssistantList(BaseList):
    def __init__(self, window=None, id=None):
        """
        Presets select menu

        :param window: main window
        :param id: input id
        """
        super(AssistantList, self).__init__(window)
        self.window = window
        self.id = id
        self.doubleClicked.connect(self.dblclick)
        self.setContextMenuPolicy(Qt.CustomContextMenu)
        self.customContextMenuRequested.connect(self.show_context_menu)
        self.restore_after_ctx_menu = True
        self._backup_selection = None

    def click(self, val):
        self.window.controller.assistant.select(val.row())

    def dblclick(self, val):
        """
        Double click event

        :param val: double click event
        """
        self.window.controller.assistant.editor.edit(val.row())

    def show_context_menu(self, pos: QPoint):
        """
        Context menu event

        :param pos: QPoint
        """
        global_pos = self.viewport().mapToGlobal(pos)
        item = self.indexAt(pos)

        actions = {}
        actions['edit'] = QAction(QIcon(":/icons/edit.svg"), trans('assistant.action.edit'), self)
        actions['edit'].triggered.connect(
            lambda checked=False, item=item: self.action_edit(item))

        actions['delete'] = QAction(QIcon(":/icons/delete.svg"), trans('assistant.action.delete'), self)
        actions['delete'].triggered.connect(
            lambda checked=False, item=item: self.action_delete(item))

        menu = QMenu(self)
        menu.addAction(actions['edit'])
        menu.addAction(actions['delete'])

        idx = item.row()
        if idx >= 0:
            #self.window.controller.assistant.select(item.row())
            menu.exec_(global_pos)

        # store previous scroll position
        self.store_scroll_position()

        # restore selection if it was backed up
        if self.restore_after_ctx_menu:
            if self._backup_selection is not None:
                self.selectionModel().clearSelection()
                for idx in self._backup_selection:
                    self.selectionModel().select(
                        idx, QItemSelectionModel.Select | QItemSelectionModel.Rows
                    )
                self._backup_selection = None

        # restore scroll position
        self.restore_after_ctx_menu = True
        self.restore_scroll_position()

    def action_edit(self, item):
        """
        Edit action handler

        :param item: list item
        """
        idx = item.row()
        if idx >= 0:
            self.restore_after_ctx_menu = False  # do not restore selection after context menu
            self.window.controller.assistant.editor.edit(idx)

    def action_delete(self, item):
        """
        Delete action handler

        :param item: list item
        """
        idx = item.row()
        if idx >= 0:
            self.restore_after_ctx_menu = False  # do not restore selection after context menu
            self.window.controller.assistant.delete(idx)

    def mousePressEvent(self, event):
        if event.button() == Qt.LeftButton:
            index = self.indexAt(event.pos())
            if not index.isValid():
                return
            super().mousePressEvent(event)
        elif event.button() == Qt.RightButton:
            index = self.indexAt(event.pos())
            if index.isValid():
                self._backup_selection = list(self.selectionModel().selectedIndexes())
                self.selectionModel().clearSelection()
                self.selectionModel().select(
                    index, QItemSelectionModel.Select | QItemSelectionModel.Rows
                )
            event.accept()
        else:
            super().mousePressEvent(event)

    def selectionCommand(self, index, event=None):
        """
        Selection command
        :param index: Index
        :param event: Event
        """
        return super().selectionCommand(index, event)



#!/usr/bin/env python3
# -*- coding: utf-8 -*-
# ================================================== #
# This file is a part of PYGPT package               #
# Website: https://pygpt.net                         #
# GitHub:  https://github.com/szczyglis-dev/py-gpt   #
# MIT License                                        #
# Created By  : Marcin Szczygliński                  #
# Updated Date: 2025.07.01 01:00:00                  #
# ================================================== #

import json
import time
from typing import Optional, Dict, Any, List

from pygpt_net.core.types import (
    MODE_CHAT,
    MODE_VISION,
    MODE_AUDIO,
    OPENAI_DISABLE_TOOLS,
    OPENAI_REMOTE_TOOL_DISABLE_CODE_INTERPRETER,
    OPENAI_REMOTE_TOOL_DISABLE_IMAGE,
    OPENAI_REMOTE_TOOL_DISABLE_WEB_SEARCH,
)
from pygpt_net.core.bridge.context import BridgeContext, MultimodalContext
from pygpt_net.item.ctx import CtxItem
from pygpt_net.item.model import ModelItem

from pygpt_net.item.attachment import AttachmentItem


class Responses:
    def __init__(self, window=None):
        """
        Responses API wrapper

        :param window: Window instance
        """
        self.window = window
        self.input_tokens = 0
        self.audio_prev_id = None
        self.audio_prev_expires_ts = None
        self.prev_response_id = None

    def send(
            self,
            context: BridgeContext,
            extra: Optional[Dict[str, Any]] = None
    ):
        """
        Call OpenAI API for chat

        :param context: Bridge context
        :param extra: Extra arguments
        :return: response or stream chunks
        """
        prompt = context.prompt
        stream = context.stream
        max_tokens = int(context.max_tokens or 0)
        system_prompt = context.system_prompt
        mode = context.mode
        model = context.model
        functions = context.external_functions
        attachments = context.attachments
        multimodal_ctx = context.multimodal_ctx

        ctx = context.ctx
        if ctx is None:
            ctx = CtxItem()  # create empty context
        user_name = ctx.input_name  # from ctx
        ai_name = ctx.output_name  # from ctx

        client = self.window.core.gpt.get_client(mode, model)

        # build chat messages
        messages = self.build(
            prompt=prompt,
            system_prompt=system_prompt,
            model=model,
            history=context.history,
            attachments=attachments,
            ai_name=ai_name,
            user_name=user_name,
            multimodal_ctx=multimodal_ctx,
        )

        msg_tokens = self.window.core.tokens.from_messages(
            messages,
            model.id,
        )
        # check if max tokens not exceeded
        if max_tokens > 0:
            if msg_tokens + int(max_tokens) > model.ctx:
                max_tokens = model.ctx - msg_tokens - 1
                if max_tokens < 0:
                    max_tokens = 0

        # extra API kwargs
        response_kwargs = {}

        # tools / functions
        tools = []
        if functions is not None and isinstance(functions, list):
            for function in functions:
                if str(function['name']).strip() == '' or function['name'] is None:
                    continue
                params = {}
                if function['params'] is not None and function['params'] != "":
                    params = json.loads(function['params'])  # unpack JSON from string
                tools.append({
                    "type": "function",
                    "name": function['name'],
                    "parameters": params,
                    "description": function['desc'],
                })

        # extra arguments, o3 only
        if model.extra and "reasoning_effort" in model.extra:
            response_kwargs['reasoning'] = {}
            response_kwargs['reasoning']['effort'] = model.extra["reasoning_effort"]

        if model.id in OPENAI_DISABLE_TOOLS:
            tools = []  # disable tools for specific models

        # extend tools with external tools
        if (not model.id.startswith("o1")
                and not model.id.startswith("o3")):  # o1, o3, o4 models do not support tools

            if not model.id in OPENAI_REMOTE_TOOL_DISABLE_WEB_SEARCH:
                if self.window.core.config.get("remote_tools.web_search", False):
                    tools.append({"type": "web_search_preview"})

            if not model.id in OPENAI_REMOTE_TOOL_DISABLE_CODE_INTERPRETER:
                if self.window.core.config.get("remote_tools.code_interpreter", False):
                    tools.append({
                        "type": "code_interpreter",
                        "container": {
                            "type": "auto"
                        }
                    })
            if not model.id in OPENAI_REMOTE_TOOL_DISABLE_IMAGE:
                if self.window.core.config.get("remote_tools.image", False):
                    tool = {"type": "image_generation"}
                    if stream:
                        tool["partial_images"] = 1  # required for streaming
                    tools.append(tool)

        # tool calls are not supported for o1-mini and o1-preview
        if (model.id is not None
                and model.id not in ["o1-mini", "o1-preview"]):
            if len(tools) > 0:
                response_kwargs['tools'] = tools

        # attach previous response ID if available
        if self.prev_response_id:
            response_kwargs['previous_response_id'] = self.prev_response_id

        response = client.responses.create(
            input=messages,
            model=model.id,
            stream=stream,
            **response_kwargs,
        )

        # store previous response ID
        if not stream and response:
            ctx.msg_id = response.id

        return response

    def build(
            self,
            prompt: str,
            system_prompt: str,
            model: ModelItem,
            history: Optional[List[CtxItem]] = None,
            attachments: Optional[Dict[str, AttachmentItem]] = None,
            ai_name: Optional[str] = None,
            user_name: Optional[str] = None,
            multimodal_ctx: Optional[MultimodalContext] = None,
    ) -> list:
        """
        Build list of chat messages

        :param prompt: user prompt
        :param system_prompt: system prompt
        :param history: history
        :param model: model item
        :param attachments: attachments
        :param ai_name: AI name
        :param user_name: username
        :param multimodal_ctx: Multimodal context
        :return: messages list
        """
        messages = []
        self.prev_response_id = None  # reset
        is_tool_output = False  # reset

        # tokens config
        mode = MODE_CHAT
        tool_call_native_enabled = self.window.core.config.get('func_call.native', False)
        allowed_system = True
        if (model.id is not None
                and model.id in ["o1-mini", "o1-preview"]):
            allowed_system = False

        used_tokens = self.window.core.tokens.from_user(
            prompt,
            system_prompt,
        )  # threshold and extra included
        max_ctx_tokens = self.window.core.config.get('max_total_tokens')  # max context window

        # fit to max model tokens
        if max_ctx_tokens > model.ctx:
            max_ctx_tokens = model.ctx

        # input tokens: reset
        self.reset_tokens()

        # append system prompt
        if allowed_system:
            if system_prompt is not None and system_prompt != "":
                messages.append({"role": "developer", "content": system_prompt})

        # append messages from context (memory)
        if self.window.core.config.get('use_context'):
            items = self.window.core.ctx.get_history(
                history,
                model.id,
                mode,
                used_tokens,
                max_ctx_tokens,
            )
            for item in items:
                # input
                if item.final_input is not None and item.final_input != "":
                    messages.append({
                        "role": "user",
                        "content": item.final_input,
                    })

                # output
                if item.final_output is not None and item.final_output != "":
                    msg = {
                        "role": "assistant",
                        "content": item.final_output,
                    }
                    # append previous audio ID
                    if MODE_AUDIO in model.mode:
                        if item.audio_id:
                            # at first check expires_at - expired audio throws error in API
                            current_timestamp = time.time()
                            audio_timestamp = int(item.audio_expires_ts) if item.audio_expires_ts else 0
                            if audio_timestamp and audio_timestamp > current_timestamp:
                                msg["audio"] = {
                                    "id": item.audio_id
                                }
                        elif self.audio_prev_id:
                            current_timestamp = time.time()
                            audio_timestamp = int(self.audio_prev_expires_ts) if self.audio_prev_expires_ts else 0
                            if audio_timestamp and audio_timestamp > current_timestamp:
                                msg["audio"] = {
                                    "id": self.audio_prev_id
                                }
                    messages.append(msg)

                    # ---- tool output ----
                    is_tool_output = False  # reset tool output flag
                    is_last_item = item == items[-1] if items else False
                    if is_last_item and tool_call_native_enabled and item.extra and isinstance(item.extra, dict):
                        if "tool_calls" in item.extra and isinstance(item.extra["tool_calls"], list):
                            for tool_call in item.extra["tool_calls"]:
                                if "function" in tool_call:
                                    if "call_id" not in tool_call or "name" not in tool_call["function"]:
                                        continue
                                    if tool_call["call_id"] and tool_call["function"]["name"]:
                                        if "tool_output" in item.extra and isinstance(item.extra["tool_output"], list):
                                            for tool_output in item.extra["tool_output"]:
                                                if ("cmd" in tool_output
                                                        and tool_output["cmd"] == tool_call["function"]["name"]):
                                                    msg = {
                                                        "type": "function_call_output",
                                                        "call_id": tool_call["call_id"],
                                                        "output": str(tool_output),
                                                    }
                                                    is_tool_output = True
                                                    messages.append(msg)
                                                    break
                                                elif "result" in tool_output:
                                                    # if result is present, append it as function call output
                                                    msg = {
                                                        "type": "function_call_output",
                                                        "call_id": tool_call["call_id"],
                                                        "output": str(tool_output["result"]),
                                                    }
                                                    is_tool_output = True
                                                    messages.append(msg)
                                                    break

                # --- previous message ID ---
                if (item.msg_id
                        and ((item.cmds is None or len(item.cmds) == 0) or is_tool_output)):  # if no cmds before or tool output
                    self.prev_response_id = item.msg_id  # previous response ID to use in current input

        # use vision and audio if available in current model
        if not is_tool_output:  # append current prompt only if not tool output
            content = str(prompt)
            if model.is_image_input():
                content = self.window.core.gpt.vision.build_content(
                    content=content,
                    attachments=attachments,
                    responses_api=True,
                )
            if model.is_audio_input():
                content = self.window.core.gpt.audio.build_content(
                    content=content,
                    multimodal_ctx=multimodal_ctx,
                )

            # append current prompt
            messages.append({
                "role": "user",
                "content": content,
            })

        # input tokens: update
        self.input_tokens += self.window.core.tokens.from_messages(
            messages,
            model.id,
        )
        return messages

    def reset_tokens(self):
        """Reset input tokens counter"""
        self.input_tokens = 0

    def get_used_tokens(self) -> int:
        """
        Get input tokens counter

        :return: input tokens
        """
        return self.input_tokens
